###################################################################
# Copyright 2013-2014 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

import os

import six

from paradrop.base.output import out
from paradrop.base import settings
from paradrop.lib.utils import pdos, pdosq


def getSystemConfigDir():
    base = settings.UCI_CONFIG_DIR
    pdosq.makedirs(base)
    return base


def getSystemPath(filename):
    """
    Get the path to the system configuration file.

    This function also attempts to create the configuration directory if it
    does not exist.

    Typical filenames: network, wireless, qos, firewall, dhcp, etc.
    """
    base = getSystemConfigDir()
    return os.path.join(base, filename)


def stringify(a):
    """
    Recursively convert all primitives in a data structure to strings.
    """
    if isinstance(a, six.string_types):
        return a
    elif isinstance(a, list):
        return [stringify(v) for v in a]
    elif isinstance(a, dict):
        result = {}
        for k, v in six.iteritems(a):
            result[k] = stringify(v)
        return result
    else:
        return str(a)


def getLineParts(line):
    """
    Split the UCI line into its whitespace-separated parts.

    Returns a list of strings, with apostrophes removed.
    """
    parts = line.split(" ")

    # Nothing to work with in this case. It is probably an erroneous line,
    # because they should usually have at least two parts.
    if len(parts) <= 1:
        return parts

    groups = []

    while len(parts) > 0:
        word = parts.pop(0)

        # Suppress empty words that are not enclosed in quotations.
        if len(word) == 0:
            continue

        opening_quote = None
        if word.startswith("'") or word.startswith('"'):
            opening_quote = word[0]
            word = word[1:]

        closing_quote = None
        if opening_quote is not None and word.endswith(opening_quote):
            closing_quote = word[-1]
            word = word[:-1]

        group = [word]

        # Simple case: line part is a single word.
        if opening_quote is None or opening_quote == closing_quote:
            groups.append(group)
            continue

        # Build up a multi-word part until we find the closing quote.
        # Example: config ssid 'Free WiFi' # <- Space inside quotation marks.
        while len(parts) > 0:
            word = parts.pop(0)

            if word.endswith(opening_quote):
                closing_quote = word[-1]
                word = word[:-1]

            group.append(word)

            if opening_quote == closing_quote:
                break

        groups.append(group)

    return [" ".join(g) for g in groups]


def chuteConfigsMatch(chutePre, chutePost):
    """ Takes two lists of objects, and returns whether or not they are identical."""
    # TODO - currently using a naive implementation by searching over the old configs and the new configs.
    # Could improve if slow by keep track of matched configs on both sides, deleting from search space
    # If any objects remain at the end
    # loop through all old configs, check if they each have a match in the new configs
    for c1 in chutePre:
        for c2 in chutePost:
            if (singleConfigMatches(c1, c2)):
                break
        else:
            # We got through the loop without finding a match, so return false
           return False

    for c2 in chutePost:
        for c1 in chutePre:
            if (singleConfigMatches(c1, c2)):
                break
        else:
            return False

    return True


def isMatch(a, b):
    a = stringify(a)
    b = stringify(b)
    return (a == b)


def isMatchIgnoreComments(a, b):
    import copy
    a1 = copy.deepcopy(a)
    b1 = copy.deepcopy(b)

    a1.pop('comment', None)
    b1.pop('comment', None)
    a1 = stringify(a1)
    b1 = stringify(b1)
    return (a1 == b1)


def singleConfigMatches(a, b):
    (c1, o1) = a
    (c2, o2) = b
    return isMatch(c1, c2) and isMatch(o1, o2)


def stringifyOptionValue(value):
    """
    Convert option value from in-memory representation to a suitable string.

    In particular, boolean values are converted to '0' or '1'.
    """
    if value is True:
        return '1'
    elif value is False:
        return '0'
    else:
        return str(value)


class UCIConfig:
    """
        Wrapper around the UCI configuration files.
            These files are found under /etc/config/*, and are used by OpenWrt to keep track of
            configuration for modules typically found in /etc/init.d/*

            The modules of interest and with current support are:
                - firewall
                - network
                - wireless
                - qos

            * This class should work with any UCI module but ALL others are UNTESTED!

        New configuration settings can be added to the UCI file via addConfig().

        Each UCI config file is expected to contain the following syntax:

            config keyA [valueA]
                option key1 value1
                ...
                list key2 value1
                list key2 value2
                ...
                list key3 value1
                list key3 value2

        Based on the UCI file above, the config syntax would look like the following:
            config is a list of tuples, containing 2 dict objects in each tuple:

                - tuple[0] describes the first line (config keyA [valueA])
                    {'type': keyA, 'name': valueA}
                    The value parameter is optional and if missing, then the 'name' key is also missing (rather than set to None).

                - tuple[1] describes the options associated with the settings (both 'option' and 'list' lines)
                    {'key1': 'value1', ...}

                    If a list is present, it looks like the following:
                        {
                            ...,
                            'key2': [value1, value2, ...],
                            'key3': [value1, value2, ...]
                        }

                So for the example above, the full config definition would look like:
                    C = {'type': 'keyA', 'name': 'valueA'}
                    O = {'key1': 'value1', 'key2': ['value1', 'value2'], 'key3': ['value1', 'value2']}
                    config = [(C, O)]
    """
    def __init__(self, filepath):
        self.filepath = filepath
        self.myname = os.path.basename(self.filepath)
        if (not os.path.isfile(self.filepath)):
            open(self.filepath, 'a').close()

        self.config = self.readConfig()

    def __eq__(self, o):
        if(self.filepath != o.filepath):
            return False
        if(self.myname != o.myname):
            return False

        # before parsing through each one, do simple check to make sure the number of configs is the same
        if(len(self.config) != len(o.config)):
            return False

        # Parse through config finding any differences
        oc = o.config
        for cfg in self.config:
            c, o = cfg
            # This is stupid slow, but easy, just look through all values for a match
            for cfg1 in oc:
                c1, o1 = cfg1
                if(c1 == c and o1 == o):
                    break
            else:
                # Found no match so we return None
                return False

        return True

    def __ne__(self, o):
        """Override the not equals operator between 2 Config objects
            This is required because the config attribute contains a list of tuples which Python doesn't
            seem to like to do comparisons directly on, for instance cfg1.config != cfg2.config fails to
            say they are the same even though they are."""
        if(self.filepath != o.filepath):
            return True
        if(self.myname != o.myname):
            return True

        # before parsing through each one, do simple check to make sure the number of configs is the same
        if(len(self.config) != len(o.config)):
            return True

        # Parse through config finding any differences
        oc = o.config
        for cfg in self.config:
            c, o = cfg
            # This is stupid slow, but easy, just look through all values for a match
            for cfg1 in oc:
                c1, o1 = cfg1
                if(c1 == c and o1 == o):
                    break
            else:
                # Found no match so we return None
                return True

        return False

    def getConfig(self, config):
        """ Returns a list of call configs with the given title """
        matches = []
        # Search through the config array for matches
        config = stringify(config)
        for e in self.config:
            c, o = e
            if(c == config):
                matches.append((c, o))

        return matches

    def getConfigIgnoreComments(self, config):
        """ Returns a list of call configs with the given title.
            Comments are ignored.
         """
        matches = []
        # Search through the config array for matches
        for e in self.config:
            c, o = e
            if(isMatchIgnoreComments(c, config)):
                matches.append((c, o))

        return matches

    def existsConfig(self, config, options):
        """Tests if the (config, options) is in the current config file."""
        # Search through the config array for matches
        config = stringify(config)
        options = stringify(options)
        for e in self.config:
            c, o = e
            if(c == config and o == options):
                return True
        return False

    def addConfigs(self, configs):
        """ Adds a list of tuples to our config """
        for e in configs:
            c, o = e
            self.addConfig(c, o)

    def delConfigs(self, configs):
        """ Adds a list of tuples to our config """
        for e in configs:
            c, o = e
            self.delConfig(c, o)

    def addConfig(self, config, options):
        """Adds the tuple to our config."""
        if (not self.existsConfig(config, options)):
            self.config.append((config, options))

    def delConfig(self, config, options):
        """Finds a match to the config input and removes it from the internal config data structure."""
        config = stringify(config)
        options = stringify(options)

        # Search through the config array for matches
        for i, e in enumerate(self.config):
            c, o = e
            if(c == config and o == options):
                break
        else:
            # Getting here means we didn't break so no match
            out.verbose('No match to delete, config: %r\n' % (config))
            return

        del(self.config[i])

    def backup(self, backupToken):
        """
            Puts a backup of this config to the location specified in @backupPath.
        """
        pdosq.makedirs(settings.UCI_BACKUP_DIR)
        backupPath = "{}/{}-{}".format(settings.UCI_BACKUP_DIR, self.myname,
                backupToken)
        pdos.copy(self.filepath, backupPath)

    def restore(self, backupToken, saveBackup=True):
        """
            Replaces real file (at /etc/config/*) with backup copy from /tmp/*-@backupToken location.

            Arguments:
                backupToken: The backup token appended at the end of the backup path
                saveBackup : A flag to keep a backup copy or delete it (default is keep backup)
        """
        # Make sure it exists!
        backupPath = "{}/{}-{}".format(settings.UCI_BACKUP_DIR, self.myname,
                backupToken)
        if(pdos.exists(backupPath)):
            if(saveBackup):
                pdos.copy(backupPath, self.filepath)
            else:
                pdos.move(backupPath, self.filepath)
        else:
            # This might be ok if they didn't actually make any changes
            out.warn('Cannot restore, %s missing backup (might be OK if no changes made)\n' % (self.myname))

    def getChuteConfigs(self, internalid):
        chuteConfigs = []
        for e in self.config:
            c, o = e
            if (c.get('comment', '') == internalid):
                chuteConfigs.append((c,o))
        return chuteConfigs

    def save(self, backupToken="paradrop", internalid=None):
        """
            Saves out the file in the proper format.

            Arguments:
                [backupPath] : Save a backup copy of the UCI file to the path provided.
                                Should be a token name like 'backup', it gets appended with a hyphen.
        """
        # Save original copy
        if(backupToken):
            self.backup(backupToken)

        output = ""
        output += "#" * 80 + "\n"
        output += "# Configuration file generated by paradrop-daemon\n".format(self.myname)
        output += "# Path: {}\n".format(self.filepath)
        output += "# Package: {}\n".format(self.myname)
        output += "#" * 80 + "\n"
        output += "\n"

        # Now generate what the file would look like
        for c, o in self.config:
            line = "config %s" % c['type']
            # Check for optional name
            if 'name' in c:
                line += " %s" % c['name']
            if 'comment' in c:
                line += " #%s" % c['comment']
            output += "%s\n" % line

            for k, v in six.iteritems(o):
                if isinstance(v, list):
                    # For list-valued options, emit one line for each item
                    # in the list.
                    for vals in v:
                        # Now append a list set to the config
                        line = "\tlist %s '%s'\n" % (k, vals)
                        output += line

                # Skip options that are None rather than writing "None".
                elif v is not None:
                    sv = stringifyOptionValue(v)
                    line = "\toption %s '%s'\n" % (k, sv)
                    output += line

            # Now add one extra newline before the next set
            output += "\n"

        # Now write to disk
        try:
            out.info('Saving %s to disk\n' % (self.filepath))
            fd = pdos.open(self.filepath, 'w')
            fd.write(output)

            # Guarantee that its written to disk before we close
            fd.flush()
            os.fsync(fd.fileno())
            fd.close()
        except Exception as e:
            out.err('Unable to save new config %s, %s\n' % (self.filepath, str(e)))
            out.err('Config may be corrupted, backup exists at /tmp/%s\n' % (self.myname))


    def readConfig(self):
        """Reads in the config file."""
        lines = []
        try:
            fd = pdos.open(self.filepath, 'r')

            while(True):
                line = fd.readline()
                if(not line):
                    break
                lines.append(line)
            fd.close()
        except Exception as e:
            out.err('Error reading file %s: %s\n' % (self.filepath, str(e)))
            raise e

        cfg = None
        opt = None
        data = []

        # Now we have the data, deal with it
        for line in lines:
            line = line.strip()

            # If comment ignore
            if line.startswith('#'):
                continue

            l = getLineParts(line)

            #
            # Config
            #
            if(l[0] == 'config'):
                # Save last config we had
                if(cfg and opt):
                    data.append((cfg, opt))

                # start a new config
                cfg = {'type': l[1]}

                # Third element can be comment or name
                if(len(l) == 3):
                    if (l[2].startswith('#')):
                        cfg['comment'] = l[2][1:]
                    else:
                        cfg['name'] = l[2]
                elif (len(l) == 4):
                    # Four elements, so third is name and 4th is comment
                        cfg['name'] = l[2]
                        cfg['comment'] = l[3][1:]
                opt = {}

            #
            # Options
            #
            elif(l[0] == 'option'):
                opt[l[1]] = l[2]

            #
            # List
            #
            elif(l[0] == 'list'):
                # Make sure the key exists and is a list.
                if l[1] not in opt:
                    opt[l[1]] = []
                elif not isinstance(opt[l[1]], list):
                    # One line started with "option", another with "list".  If
                    # this is supposed to be a list, they should all start with
                    # "list".
                    raise Exception("Malformed UCI: mixed list/option lines")

                opt[l[1]].append(l[2])

        else:
            # Also at the end of the loop, save the final config we were making
            # Make sure cfg,opt aren't None
            if(None not in (cfg, opt)):
                data.append((cfg, opt))

        return data



from . import main


def reloadAll():
    """
    Reload all files from the system configuration directory.

    This function blocks until the request completes.  On completion it returns
    a status string, which is a JSON list of loaded configuration sections with
    a 'success' field.
    For critical errors it will return None.
    """
    return main.configManager.loadConfig()


def reload(path):
    """
    Reload file(s) specified by path.

    This function blocks until the request completes.  On completion it returns
    a status string, which is a JSON list of loaded configuration sections with
    a 'success' field.
    For critical errors it will return None.
    """
    return main.configManager.loadConfig(path)


def systemStatus():
    """
    Return system status string from pdconf.
    """
    return main.configManager.statusString()


def waitSystemUp():
    """
    Wait for the configuration daemon to finish its first load.

    This function blocks until the request completes.  On completion it returns
    a status string, which is a JSON list of loaded configuration sections with
    a 'success' field.
    For critical errors it will return None.
    """
    return main.configManager.waitSystemUp()



'''
Core module.
Contains the entry point into Paradrop and establishes all other modules.
Does not implement any behavior itself.
'''

import argparse
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks

from paradrop.base.output import out
from paradrop.base import nexus, settings
from paradrop.lib.misc.procmon import ProcessMonitor
from paradrop.core.agent import provisioning
from paradrop.core.agent.reporting import sendNodeIdentity, sendStateReport
from paradrop.core.agent.wamp_session import WampSession
from paradrop.core.update.update_fetcher import UpdateFetcher
from paradrop.core.update.update_manager import UpdateManager
from paradrop.airshark.airshark import AirsharkManager
from paradrop.backend.http_server import HttpServer, setup_http_server
from paradrop.backend.ssdp_responder import SsdpResponder
from paradrop import confd


class Nexus(nexus.NexusBase):
    def __init__(self, update_fetcher, update_manager):
        self.update_fetcher = update_fetcher
        self.update_manager = update_manager
        # Want to change logging functionality? See optional args on the base class and pass them here
        super(Nexus, self).__init__(stealStdio=True, printToConsole=True)


    @inlineCallbacks
    def onStart(self):
        super(Nexus, self).onStart()
        # onStart is called when the reactor starts, not when the connection is made.
        # Check for provisioning keys and attempt to connect
        if not self.provisioned() and not provisioning.can_provision():
            out.warn("The node is not provisioned and is not configured to self-provision.")
            return

        while not self.provisioned():
            yield provisioning.provision_self(self.update_manager)

        try:
            # Set up communication with pdserver.
            # 1. Create a report of the current system state and send that.
            # 2. Send the node public key.
            # 3. Poll for a list of updates that should be applied.
            # 4. Open WAMP session.
            yield sendStateReport()
            yield sendNodeIdentity()
            yield self.update_fetcher.start_polling()
            yield self.connect(WampSession)
        except Exception:
            out.warn('The router ID or password is invalid!')
        else:
            out.info('WAMP session is ready!')

    def onStop(self):
        if self.session is not None:
            self.session.leave()
        else:
            out.info('No WAMP session found!')

        super(Nexus, self).onStop()


def main():
    p = argparse.ArgumentParser(description='Paradrop daemon running on client')
    p.add_argument('--mode', '-m', help='Set the mode to one of [production, local, unittest]',
                   action='store', type=str, default='production')
    p.add_argument('--portal', '-p', help='Set the folder of files for local portal',
                   action='store', type=str)
    p.add_argument('--no-exec', help='Skip execution of configuration commands',
                   action='store_false', dest='execute')

    args = p.parse_args()
    # print args

    settings.loadSettings(args.mode, [])

    update_manager = UpdateManager(reactor)
    update_fetcher = UpdateFetcher(update_manager)
    WampSession.set_update_fetcher(update_fetcher)
    ProcessMonitor.allowedActions = set()

    # Start the configuration service as a thread
    confd.main.run_thread(execute=args.execute)

    airshark_manager = AirsharkManager()

    # Globally assign the nexus object so anyone else can access it.
    nexus.core = Nexus(update_fetcher, update_manager)
    http_server = HttpServer(update_manager, update_fetcher, airshark_manager, args.portal)
    setup_http_server(http_server, '0.0.0.0', settings.PORTAL_SERVER_PORT)
    reactor.listenMulticast(1900, SsdpResponder(), listenMultiple=True)

    reactor.run()


if __name__ == "__main__":
    main()



'''
Provide information of the router, e.g. board version, CPU information,
memory size, disk size.

Endpoints for these functions can be found under /api/v1/info.
'''
import json
import os
import platform
from psutil import virtual_memory
from klein import Klein

from paradrop.base import constants
from paradrop.core.config.devices import detectSystemDevices
from paradrop.core.system.system_info import getOSVersion, getPackageVersion
from paradrop.core.agent.reporting import TelemetryReportBuilder
from paradrop.lib.utils import pdos
from . import cors


class InformationApi(object):
    routes = Klein()

    def __init__(self):
        self.vendor = pdos.read_sys_file('/sys/devices/virtual/dmi/id/sys_vendor')
        self.board = pdos.read_sys_file('/sys/devices/virtual/dmi/id/product_name', default='') \
                     + ' ' + pdos.read_sys_file('/sys/devices/virtual/dmi/id/product_version', default='')
        self.cpu = platform.processor()
        self.memory = virtual_memory().total

        self.wifi = []
        devices = detectSystemDevices()
        for wifiDev in devices['wifi']:
            self.wifi.append({
                'id': wifiDev['id'],
                'macAddr': wifiDev['mac'],
                'vendorId': wifiDev['vendor'],
                'deviceId': wifiDev['device'],
                'slot': wifiDev['slot']
            })

        self.biosVendor = pdos.read_sys_file('/sys/devices/virtual/dmi/id/bios_vendor')
        self.biosVersion = pdos.read_sys_file('/sys/devices/virtual/dmi/id/bios_version')
        self.biosDate = pdos.read_sys_file('/sys/devices/virtual/dmi/id/bios_date')
        self.osVersion = getOSVersion()
        self.kernelVersion = platform.system() + '-' + platform.release()
        self.pdVersion = getPackageVersion('paradrop')
        self.uptime = int(float(pdos.read_sys_file('/proc/uptime', default='0').split()[0]))

    @routes.route('/hardware', methods=['GET'])
    def hardware_info(self, request):
        """
        Get information about the hardware platform.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/info/hardware

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "wifi": [
               {
                 "slot": "pci/0000:04:00.0",
                 "vendorId": "0x168c",
                 "macAddr": "04:f0:21:2f:b7:c1",
                 "id": "pci-wifi-0",
                 "deviceId": "0x003c"
               },
               {
                 "slot": "pci/0000:06:00.0",
                 "vendorId": "0x168c",
                 "macAddr": "04:f0:21:0f:78:28",
                 "id": "pci-wifi-1",
                 "deviceId": "0x002a"
               }
             ],
             "memory": 2065195008,
             "vendor": "PC Engines",
             "board": "APU 1.0",
             "cpu": "x86_64"
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')
        data = dict()
        data['vendor'] = self.vendor
        data['board'] = self.board
        data['cpu'] = self.cpu
        data['memory'] = self.memory
        data['wifi'] = self.wifi
        return json.dumps(data)

    @routes.route('/software', methods=['GET'])
    def software_info(self, request):
        """
        Get information about the operating system.

        Returns a dictionary containing information the BIOS version, OS
        version, kernel version, Paradrop version, and system uptime.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/info/software

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "biosVersion": "SageBios_PCEngines_APU-45",
             "biosDate": "04/05/2014",
             "uptime": 15351,
             "kernelVersion": "Linux-4.4.0-101-generic",
             "pdVersion": "0.9.2",
             "biosVendor": "coreboot",
             "osVersion": "Ubuntu 4.4.0-101.124-generic 4.4.95"
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')
        data = dict()
        data['biosVendor'] = self.biosVendor
        data['biosVersion'] = self.biosVersion
        data['biosDate'] = self.biosDate
        data['kernelVersion'] = self.kernelVersion
        data['osVersion'] = self.osVersion
        data['pdVersion'] = self.pdVersion
        data['uptime'] = self.uptime
        return json.dumps(data)

    @routes.route('/environment', methods=['GET'])
    def get_environment(self, request):
        """
        Get environment variables.

        Returns a dictionary containing the environment variables passed to the
        Paradrop daemon.  This is useful for development and debugging purposes
        (e.g. see how PATH is set on Paradrop when running in different
        contexts).

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/info/environment

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "LANG": "C.UTF-8",
             "SNAP_REVISION": "x73",
             "SNAP_COMMON": "/var/snap/paradrop-daemon/common",
             "XDG_RUNTIME_DIR": "/run/user/0/snap.paradrop-daemon",
             "SNAP_USER_COMMON": "/root/snap/paradrop-daemon/common",
             "SNAP_LIBRARY_PATH": "/var/lib/snapd/lib/gl:/var/lib/snapd/void",
             "SNAP_NAME": "paradrop-daemon",
             "PWD": "/var/snap/paradrop-daemon/x73",
             "PATH": "/snap/paradrop-daemon/x73/usr/sbin:/snap/paradrop-daemon/x73/usr/bin:/snap/paradrop-daemon/x73/sbin:/snap/paradrop-daemon/x73/bin:/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin:/usr/games:/usr/local/games",
             "SNAP": "/snap/paradrop-daemon/x73",
             "SNAP_DATA": "/var/snap/paradrop-daemon/x73",
             "SNAP_VERSION": "0.9.2",
             "SNAP_ARCH": "amd64",
             "SNAP_USER_DATA": "/root/snap/paradrop-daemon/x73",
             "TEMPDIR": "/tmp",
             "HOME": "/root/snap/paradrop-daemon/x73",
             "SNAP_REEXEC": "",
             "LD_LIBRARY_PATH": "/var/lib/snapd/lib/gl:/var/lib/snapd/void:/snap/paradrop-daemon/x73/usr/lib/x86_64-linux-gnu::/snap/paradrop-daemon/x73/lib:/snap/paradrop-daemon/x73/usr/lib:/snap/paradrop-daemon/x73/lib/x86_64-linux-gnu:/snap/paradrop-daemon/x73/usr/lib/x86_64-linux-gnu",
             "TMPDIR": "/tmp"
             ...
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')
        return json.dumps(os.environ.data)

    @routes.route('/features', methods=['GET'])
    def get_features(self, request):
        """
        Get features supported by the host.

        This is a list of strings specifying features supported by the daemon.

        **Explanation of feature strings**:

        hostapd-control
          The daemon supports the hostapd control interface and
          provides a websocket channel for accessing it.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/info/features

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           [
             "hostapd-control"
           ]
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')
        features = constants.DAEMON_FEATURES.split()
        return json.dumps(features)

    @routes.route('/telemetry', methods=['GET'])
    def get_telemetry(self, request):
        """
        Get a telemetry report.

        This contains information about resource utilization by chute and
        system totals.  This endpoint returns the same data that we
        periodically send to the controller if telemetry is enabled.
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')
        builder = TelemetryReportBuilder()
        report = builder.prepare()
        return json.dumps(report)



'''
Write the CROSS-ORIGIN RESOURCE SHARING headers required
Reference: http://msoulier.wordpress.com/2010/06/05/cross-origin-requests-in-twisted/
'''


def config_cors(request):
    request.setHeader('Access-Control-Allow-Origin', '*')
    request.setHeader('Access-Control-Allow-Methods', 'GET, PUT, POST, DELETE')
    request.setHeader('Access-Control-Allow-Headers', 'Content-Type')
    request.setHeader('Access-Control-Max-Age', 2520) # 42 hours



import os

from twisted.internet import reactor
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET

from paradrop.base import settings
from paradrop.lib.utils.uhttp import UHTTPConnection
from . import cors

class SnapdResource(Resource):
    """
    Expose the snapd API by forwarding requests.

    Changed in 0.13: we try to send the request through the governor service so
    that paradrop can be installed in strict mode.

    https://github.com/snapcore/snapd/wiki/REST-API
    """

    isLeaf = True

    def do_snapd_request(self, request):
        """
        Forward the API request to snapd.
        """
        if os.path.exists(settings.GOVERNOR_INTERFACE):
            interface = settings.GOVERNOR_INTERFACE
            path = '/snapd/' + '/'.join(request.postpath)
        else:
            interface = settings.SNAPD_INTERFACE
            path = '/' + '/'.join(request.postpath)

        body = None
        headers = {}

        # Try to read Content-Type of original request, and set up the body of
        # the request to snapd.  We should expect None for GET requests and
        # JSON for POST requests.
        ctype = request.requestHeaders.getRawHeaders('Content-Type',
                default=[None])[0]
        if ctype is not None:
            headers['Content-Type'] = ctype
            body = request.content.read()

        # Make the request to snapd using a Unix socket.
        conn = UHTTPConnection(interface)
        conn.request(request.method, path, body, headers)
        res = conn.getresponse()
        data = res.read()

        # Return the response to the original request.
        request.setHeader('Content-Type', 'application/json')
        request.write(data)
        request.finish()

    def render(self, request):
        """
        Fulfill requests by forwarding them to snapd.

        We use a synchronous implementation of HTTP over Unix sockets, so we do
        the request in a worker thread and have it call request.finish.
        """
        cors.config_cors(request)
        reactor.callInThread(self.do_snapd_request, request)
        return NOT_DONE_YET



import subprocess


def reboot(update):
    """
    Reboot the node.
    """
    cmd = ["shutdown", "-r", "+1"]
    result = subprocess.call(cmd)
    if result == 0:
        update.progress("Rebooting in 60 seconds.")
    else:
        raise Exception("shutdown command failed, returned {}".format(result))


def shutdown(update):
    """
    Power down the node.
    """
    cmd = ["shutdown", "-h", "+1"]
    result = subprocess.call(cmd)
    if result == 0:
        update.progress("Halting in 60 seconds.")
    else:
        raise Exception("shutdown command failed, returned {}".format(result))



from autobahn.twisted.websocket import WebSocketServerProtocol
from autobahn.twisted.websocket import WebSocketServerFactory
import smokesignal

from paradrop.base.output import out

class ParadropLogWsProtocol(WebSocketServerProtocol):
    def __init__(self, factory):
        WebSocketServerProtocol.__init__(self)
        self.factory = factory

    def onOpen(self):
        out.info('ws /paradrop_logs connected')
        self.factory.addParadropLogObserver(self)

    def onParadropLog(self, logDict):
        self.sendMessage(logDict['message'])

    def onClose(self, wasClean, code, reason):
        out.info('ws /paradrop_logs disconnected: {}'.format(reason))
        self.factory.removeParadropLogObserver(self)


class ParadropLogWsFactory(WebSocketServerFactory):
    def __init__(self, *args, **kwargs):
        WebSocketServerFactory.__init__(self, *args, **kwargs)
        self.observers = []

    def buildProtocol(self, addr):
        return ParadropLogWsProtocol(self)

    def addParadropLogObserver(self, observer):
        if (self.observers.count(observer) == 0):
            self.observers.append(observer)
            if len(self.observers) == 1:
                smokesignal.on('logs', self.onParadropLog)

    def removeParadropLogObserver(self, observer):
        if (self.observers.count(observer) == 1):
            self.observers.remove(observer)
            if len(self.observers) == 0:
                smokesignal.disconnect(self.onParadropLog)

    def onParadropLog(self, logDict):
        for observer in self.observers:
            observer.onParadropLog(logDict)



###################################################################
# Copyright 2013-2015 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

from paradrop.base.output import out
from paradrop.core.config import firewall, haproxy

from . import plangraph


def generatePlans(update):
    """
        This function looks at a diff of the current Chute (in @chuteStor) and the @newChute,
        then adds Plan() calls to make the Chute match the @newChute.

        Returns:
            True: abort the plan generation process
    """
    out.header("%r\n" % (update))

    # Make sure we need to create this chute (does it already exist)
    # TODO
#    chutePlan.addPlans(new, plangraph.TRAFFIC_SECURITY_CHECK, (security.checkTraffic, (chuteStor, new)))

    # First time, so generate the basic firewall rules in cache (key: 'osFirewallConfig')
    update.plans.addPlans(plangraph.TRAFFIC_GET_OS_FIREWALL, (firewall.getOSFirewallRules, ))

    # Get developer firewall rules (key: 'developerFirewallRules')
    update.plans.addPlans(plangraph.TRAFFIC_GET_DEVELOPER_FIREWALL, (firewall.getDeveloperFirewallRules, ))

    # Combine rules from above two fucntions, save to file
    todoPlan = (firewall.setOSFirewallRules, )
    abtPlan = (firewall.revert_os_firewall_rules, )
    update.plans.addPlans(plangraph.TRAFFIC_SET_OS_FIREWALL, todoPlan, abtPlan)

    update.plans.addPlans(plangraph.RECONFIGURE_PROXY,
            (haproxy.reconfigureProxy, ))

    return None



"""
This module listens for messages and triggers reloading of configuration files.
This module is the service side of the implementation.  If you want to
issue reload commands to the service, see the client.py file instead.

"""

from twisted.internet import reactor
from paradrop.base import settings

from .manager import ConfigManager

configManager = None

def listen(configManager):
    # Things get messy if pdconfd is restarted with running chutes.  Then it
    # will try to reconfigure the system.  One easy solution is to unload the
    # configuration before exiting.
    reactor.addSystemEventTrigger('before', 'shutdown',
                                  configManager.unload)

    # Now load all of the configuration for the first time.
    configManager.loadConfig()

def run_thread(execute=True):
    """
    Start pdconfd service as a thread.

    This function schedules pdconfd to run as a thread and returns immediately.
    """
    global configManager
    configManager = ConfigManager(settings.PDCONFD_WRITE_DIR, execute)
    reactor.callFromThread(listen, configManager)



"""
Configure optional additional services such as telemetry.
"""
from paradrop.core.agent import reporting
from paradrop.lib.utils import datastruct

from twisted.internet.task import LoopingCall


telemetry_looping_call = None


def configure_telemetry(update):
    global telemetry_looping_call

    hostConfig = update.cache_get('hostConfig')

    enabled = datastruct.getValue(hostConfig, 'telemetry.enabled', False)
    interval = datastruct.getValue(hostConfig, 'telemetry.interval', 60)

    # Cancel the old looping call.
    if telemetry_looping_call is not None:
        telemetry_looping_call.stop()
        telemetry_looping_call = None

    if enabled and interval > 0:
        telemetry_looping_call = LoopingCall(reporting.sendTelemetryReport)
        telemetry_looping_call.start(interval, now=False)



import json

from autobahn.twisted.websocket import WebSocketServerProtocol
from autobahn.twisted.websocket import WebSocketServerFactory


class ChangeStreamProtocol(WebSocketServerProtocol):
    def __init__(self, factory):
        WebSocketServerProtocol.__init__(self)
        self.factory = factory

    def onOpen(self):
        self.factory.update.add_message_observer(self)

    def on_complete(self):
        self.sendClose()

    def on_message(self, data):
        self.sendMessage(json.dumps(data))

    def onClose(self, wasClean, code, reason):
        self.factory.update.remove_message_observer(self)


class ChangeStreamFactory(WebSocketServerFactory):
    def __init__(self, update, *args, **kwargs):
        WebSocketServerFactory.__init__(self, *args, **kwargs)
        self.update = update

    def buildProtocol(self, addr):
        return ChangeStreamProtocol(self)



"""
Module for checking resource reservations by chutes.

One idea motivating this design is to reduce the amount of state
in memory for resource reservations.  We have the chute list, which
contains information about what devices the chute is using.  If we also
maintain a separate list of devices used by chutes, we need to keep
them synchronized.  This becomes messy when a chute fails to install
or uninstall correctly.  The getDeviceReservations function iterates
over the chute list and returns an up-to-date view of device usage.
This can be called as needed.
"""
import collections
import ipaddress


from paradrop.base import constants
from paradrop.core.config.devices import getWirelessPhyName
from paradrop.core.config.hostconfig import prepareHostConfig
from paradrop.core.chute.chute_storage import ChuteStorage
from paradrop.lib.utils import datastruct


class DeviceReservations(object):
    def __init__(self):
        self.reservations = []

    def add(self, chute, dtype, mode=None):
        r = {
            'chute': chute,
            'type': dtype,
            'mode': mode
        }

        self.reservations.append(r)

    def count(self, dtype=None, mode=None):
        """
        Return the number of reservations matching the given criteria.

        None is used as a wildcard, so if no arguments are passed, the count
        returned is the total number of reservations.
        """
        count = 0
        for res in self.reservations:
            if dtype is not None and dtype != res['type']:
                continue
            if mode is not None and mode != res['mode']:
                continue
            count += 1
        return count


def getDeviceReservations(exclude=None):
    """
    Produce a dictionary mapping device names to DeviceReservations objects
    that describe the current usage of the device.

    The returned type is a defaultdict, so there is no need to check if a key
    exists before accessing it.

    exclude: name of chute whose device reservations should be excluded
    """
    reservations = collections.defaultdict(DeviceReservations)

    if exclude != constants.RESERVED_CHUTE_NAME:
        hostConfig = prepareHostConfig()

        wifiInterfaces = hostConfig.get('wifi-interfaces', [])
        for iface in wifiInterfaces:
            if 'device' not in iface:
                continue

            dev = iface['device']
            phy = getWirelessPhyName(dev)
            if phy is not None:
                # It is annoying to do this conversion everywhere, but it would
                # be painful to break compatibility with all of the devices out
                # there that use e.g. wlan0 instead of phy0 in their hostconfig.
                dev = phy

            reservations[dev].add(constants.RESERVED_CHUTE_NAME, 'wifi',
                    iface.get('mode', 'ap'))

        lanInterfaces = datastruct.getValue(hostConfig, 'lan.interfaces', [])
        for iface in lanInterfaces:
            reservations[iface].add(constants.RESERVED_CHUTE_NAME, 'lan', None)

    for chute in ChuteStorage.chuteList.values():
        if chute.name == exclude:
            continue

        for iface in chute.getCache('networkInterfaces'):
            dev = iface.get('device', None)

            # Device is not set in cases such as vlan interfaces.
            if dev is None:
                continue

            reservations[dev].add(chute.name, iface['type'],
                    iface.get('mode', None))

    return reservations


class InterfaceReservationSet(object):
    def __init__(self):
        self.reservations = set()

    def add(self, interface):
        self.reservations.add(interface)

    def __contains__(self, x):
        return x in self.reservations

    def __len__(self):
        return len(self.reservations)


def getInterfaceReservations(exclude=None):
    """
    Get current set of interface reservations.

    Returns an instance of InterfaceReservationSet.

    exclude: name of chute whose interfaces should be excluded
    """
    reservations = InterfaceReservationSet()

    if exclude != constants.RESERVED_CHUTE_NAME:
        hostConfig = prepareHostConfig()
        wifiInterfaces = hostConfig.get('wifi-interfaces', [])
        for iface in wifiInterfaces:
            if 'ifname' not in iface:
                continue

            ifname = iface['ifname']
            reservations.add(ifname)

    for chute in ChuteStorage.chuteList.values():
        if chute.name == exclude:
            continue

        for iface in chute.getCache('networkInterfaces'):
            if 'externalIntf' not in iface:
                continue

            ifname = iface['externalIntf']
            reservations.add(ifname)

    return reservations


class SubnetReservationSet(object):
    def __init__(self):
        self.reservations = []

    def add(self, subnet):
        self.reservations.append(subnet)

    def __contains__(self, subnet):
        for res in self.reservations:
            if res.overlaps(subnet):
                return True
        return False

    def __len__(self):
        return len(self.reservations)


def getSubnetReservations(exclude=None):
    """
    Get current set of subnet reservations.

    Returns an instance of SubnetReservationSet.

    exclude: name of chute whose reservations should be excluded
    """
    reservations = SubnetReservationSet()

    if exclude != constants.RESERVED_CHUTE_NAME:
        hostConfig = prepareHostConfig()
        ipaddr = datastruct.getValue(hostConfig, 'lan.ipaddr', None)
        netmask = datastruct.getValue(hostConfig, 'lan.netmask', None)
        if ipaddr is not None and netmask is not None:
            network = ipaddress.ip_network(u'{}/{}'.format(ipaddr, netmask),
                    strict=False)
            reservations.add(network)

    for chute in ChuteStorage.chuteList.values():
        if chute.name == exclude:
            continue

        for iface in chute.getCache('networkInterfaces'):
            if 'subnet' not in iface:
                continue

            subnet = iface['subnet']
            reservations.add(subnet)

    return reservations


def getReservations(update):
    """
    Get device and resource reservations claimed by other users.
    """
    devices = getDeviceReservations(exclude=update.new.name)
    interfaces = getInterfaceReservations(exclude=update.new.name)
    subnets = getSubnetReservations(exclude=update.new.name)

    update.cache_set('deviceReservations', devices)
    update.cache_set('interfaceReservations', interfaces)
    update.cache_set('subnetReservations', subnets)



###################################################################
# Copyright 2013-2015 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

from paradrop.base.output import out
from paradrop.core.config import devices, files, network, hostconfig, reservations, wifi

from . import plangraph


def generatePlans(update):
    """
        This function looks at a diff of the current Chute (in @chuteStor) and the @newChute,
        then adds Plan() calls to make the Chute match the @newChute.

        Returns:
            True: abort the plan generation process
    """
    out.verbose("%r\n" % (update))

    update.plans.addPlans(plangraph.DOWNLOAD_CHUTE_FILES,
                          (files.download_chute_files, ),
                          (files.delete_chute_files, ))

    update.plans.addPlans(plangraph.LOAD_CHUTE_CONFIGURATION,
                          (files.load_chute_configuration, ))

    # Detect system devices and set up basic configuration for them (WAN
    # interface, wireless devices).  These steps do not need to be reverted on
    # abort.
    #
    # abortNetworkConfig is added as an abort command here so that it runs when
    # config.network.getNetworkConfig or just about anything else fails.
    update.plans.addPlans(plangraph.STRUCT_GET_SYSTEM_DEVICES,
                          (devices.getSystemDevices, ),
                          (network.abortNetworkConfig, ))

    update.plans.addPlans(plangraph.STRUCT_GET_RESERVATIONS,
                          (reservations.getReservations, ))

    update.plans.addPlans(plangraph.STRUCT_GET_HOST_CONFIG,
                          (hostconfig.getHostConfig, ))
    update.plans.addPlans(plangraph.STRUCT_SET_HOST_CONFIG,
                          (hostconfig.setHostConfig, ),
                          (hostconfig.revertHostConfig, ))

    # Save current network configuration into chute cache (key: 'networkInterfaces')
    update.plans.addPlans(plangraph.STRUCT_GET_INT_NETWORK,
                          (network.getNetworkConfig, ))

    # Setup changes to push into OS config files (key: 'osNetworkConfig')
    update.plans.addPlans(plangraph.STRUCT_GET_OS_NETWORK, (network.getOSNetworkConfig, ))

    # Setup changes to push into OS config files (key: 'osWirelessConfig')
    update.plans.addPlans(plangraph.STRUCT_GET_OS_WIRELESS, (wifi.getOSWirelessConfig, ))

    # Changes for networking
    todoPlan = (network.setOSNetworkConfig, )
    abtPlan = (network.revert_os_network_config, )
    update.plans.addPlans(plangraph.STRUCT_SET_OS_NETWORK, todoPlan, abtPlan)

    # Changes for wifi
    todoPlan = (wifi.setOSWirelessConfig, )
    abtPlan = (wifi.revert_os_wireless_config, )
    update.plans.addPlans(plangraph.STRUCT_SET_OS_WIRELESS, todoPlan, abtPlan)

    # Layer 3 bridging support
    todoPlan = (network.getL3BridgeConfig, )
    update.plans.addPlans(plangraph.STRUCT_GET_L3BRIDGE_CONFIG, todoPlan)

    todoPlan = (network.setL3BridgeConfig, )
    abtPlan = (network.revert_l3_bridge_config, )
    update.plans.addPlans(plangraph.STRUCT_SET_L3BRIDGE_CONFIG, todoPlan, abtPlan)

    return None



"""
Detect physical devices that can be used by chutes.

This module detects physical devices (for now just network interfaces) that can
be used by chutes.  This includes WAN interfaces for Internet connectivity and
WiFi interfaces which can host APs.

It also makes sure certain entries exist in the system UCI files for these
devices, for example "wifi-device" sections.  These are shared between chutes,
so they only need to be added when missing.
"""

import netifaces
import operator
import os
import re
import subprocess

import six

from paradrop.base.output import out
from paradrop.base import constants, settings
from paradrop.base.exceptions import DeviceNotFoundException
from paradrop.lib.utils import datastruct, pdos, uci


IEEE80211_DIR = "/sys/class/ieee80211"
SYS_DIR = "/sys/class/net"
EXCLUDE_IFACES = set(["lo"])

# Strings that identify a virtual interface.
VIF_MARKERS = [".", "veth"]

# Matches various ways of specifying WiFi devices (phy0, wlan0).
WIFI_DEV_REF = re.compile("([a-z]+)(\d+)")

# Set of wifi-interface mode values that are handled by Paradrop rather than
# UCI configuration system.
WIFI_NONSTANDARD_MODES = set(["airshark"])


def isVirtual(ifname):
    """
    Test if an interface is a virtual one.

    FIXME: This just tests for the presence of certain strings in the interface
    name, so it is not very robust.
    """
    for marker in VIF_MARKERS:
        if marker in ifname:
            return True
    return False


def isWAN(ifname):
    """
    Test if an interface is a WAN interface.
    """
    pattern = re.compile(r"(\w+)\s+(\w+)*")
    routeList = pdos.readFile("/proc/net/route")
    for line in routeList:
        match = pattern.match(line)
        if match is not None and \
                match.group(1) == ifname and \
                match.group(2) == "00000000":
            return True
    return False


def isWireless(ifname):
    """
    Test if an interface is a wireless device.
    """
    check_path = "{}/{}/wireless".format(SYS_DIR, ifname)
    return pdos.exists(check_path)


def detectSystemDevices():
    """
    Detect devices on the system.

    The result is three lists stored in a dictionary.  The three lists are
    indexed by 'wan', 'wifi', and 'lan'.  Other devices may be supported by
    adding additional lists.

    Within each list, a device is represented by a dictionary.
    For all devices, the 'name' and 'mac' fields are defined.
    For WiFi devices, the 'phy' is defined in addition.
    Later, we may fill in more device information
    (e.g. what channels a WiFi card supports).
    """
    devices = dict()
    devices['wan'] = list()
    devices['wifi'] = list()
    devices['lan'] = list()

    for dev in listSystemDevices():
        devices[dev['type']].append(dev)
        del dev['type']

    return devices


def readSysFile(path):
    try:
        with open(path, 'r') as source:
            return source.read().strip()
    except:
        return None


def getMACAddress(ifname):
    path = "{}/{}/address".format(SYS_DIR, ifname)
    return readSysFile(path)


def getPhyMACAddress(phy):
    path = "{}/{}/macaddress".format(IEEE80211_DIR, phy)
    return readSysFile(path)


def getWirelessPhyName(ifname):
    path = "{}/{}/phy80211/name".format(SYS_DIR, ifname)
    return readSysFile(path)


class SysReader(object):
    PCI_BUS_ID = re.compile(r"\d+:\d+:\d+\.\d+")
    USB_BUS_ID = re.compile(r"\d+\-\d+(\.\d+)*:\d+\.\d+")

    def __init__(self, phy):
        self.phy = phy
        self.device_path = "{}/{}/device".format(IEEE80211_DIR, phy)

    def getDeviceId(self, default="????"):
        """
        Return the device ID for the device.

        This is a four-digit hexadecimal number.  For example, our Qualcomm
        802.11n chips have device ID 002a.
        """
        path = os.path.join(self.device_path, "device")
        device = readSysFile(path)
        if device is None:
            device = default
        return device

    def getSlotName(self, default="????"):
        """
        Return the PCI/USB slot name for the device.

        Example: "pci/0000:04:00.0" or "usb/1-1:1.0"
        """
        path = os.path.join(self.device_path, "driver")
        for fname in os.listdir(path):
            match = SysReader.PCI_BUS_ID.match(fname)
            if match is not None:
                return "pci/" + fname

            match = SysReader.USB_BUS_ID.match(fname)

            if match is not None:
                return "usb/" + fname

        return default

    def getVendorId(self, default="????"):
        """
        Return the vendor ID for the device.

        This is a four-digit hexadecimal number.  For example, our Qualcomm
        802.11n chips have vendor ID 168c.
        """
        path = os.path.join(self.device_path, "vendor")
        vendor = readSysFile(path)
        if vendor is None:
            vendor = default
        return vendor

    def read_uevent(self):
        """
        Read the device uevent file and return the contents as a dictionary.
        """
        result = dict()
        path = os.path.join(self.device_path, "uevent")
        with open(path, "r") as source:
            for line in source:
                key, value = line.split("=")
                result[key] = value
        return result


def listWiFiDevices():
    # Collect information about the physical devices (e.g. phy0 -> MAC address,
    # device type, PCI slot, etc.) and store as objects in a dictionary.
    devices = dict()
    try:
        for phy in pdos.listdir(IEEE80211_DIR):
            mac = getPhyMACAddress(phy)
            reader = SysReader(phy)

            devices[phy] = {
                'name': "wifi{}".format(mac.replace(':', '')),
                'type': 'wifi',
                'mac': mac,
                'phy': phy,
                'vendor': reader.getVendorId(),
                'device': reader.getDeviceId(),
                'slot': reader.getSlotName()
            }

    except OSError:
        # If we get an error here, it probably just means there are no WiFi
        # devices.
        pass

    # Collect a list of interfaces corresponding to each physical device
    # (e.g. phy0 -> wlan0, vwlan0.0000, etc.)
    interfaces = dict((dev, []) for dev in devices.keys())
    for ifname in pdos.listdir(SYS_DIR):
        try:
            path = "{}/{}/phy80211/name".format(SYS_DIR, ifname)
            phy = readSysFile(path)

            path = "{}/{}/ifindex".format(SYS_DIR, ifname)
            ifindex = int(readSysFile(path))

            interfaces[phy].append({
                'ifname': ifname,
                'ifindex': ifindex
            })
        except:
            # Error probably means it was not a wireless interface.
            pass

    # Sort by ifindex to identify the primary interface, which is the one that
    # was created when the device was first added.  We make use the fact that
    # Linux uses monotonically increasing ifindex values.
    for phy, device in six.iteritems(devices):
        if len(interfaces[phy]) > 0:
            interfaces[phy].sort(key=operator.itemgetter('ifindex'))
            device['primary_interface'] = interfaces[phy][0]['ifname']
        else:
            device['primary_interface'] = None

    # Finally, sort the device list by PCI/USB slot to create an ordering that
    # is stable across device reboots and somewhat stable across hardware
    # swaps.
    result = list(devices.values())
    result.sort(key=operator.itemgetter('slot'))

    pci_index = 0
    usb_index = 0
    for dev in result:
        if dev['slot'].startswith("pci"):
            dev['id'] = "pci-wifi-{}".format(pci_index)
            pci_index += 1
        elif dev['slot'].startswith("usb"):
            dev['id'] = "usb-wifi-{}".format(usb_index)
            usb_index += 1

    return result


def listSystemDevices():
    """
    Detect devices on the system.

    The result is a single list of dictionaries, each containing information
    about a network device.
    """
    devices = list()

    for ifname in pdos.listdir(SYS_DIR):
        if ifname in EXCLUDE_IFACES:
            continue

        # Only want to detect physical interfaces.
        if isVirtual(ifname):
            continue

        # More special cases to ignore for now.
        if ifname.startswith("br"):
            continue
        if ifname.startswith("docker"):
            continue
        if ifname.startswith("sit"):
            continue

        dev = {
            'name': ifname,
            'mac': getMACAddress(ifname)
        }

        if isWAN(ifname):
            dev['type'] = 'wan'
        elif isWireless(ifname):
            # Detect wireless devices separately.
            continue
        else:
            dev['type'] = 'lan'

        devices.append(dev)

    wifi_devices = listWiFiDevices()
    devices.extend(wifi_devices)

    return devices


def resetWirelessDevice(phy, primary_interface):
    """
    Reset a wireless device's interfaces to clean state.

    This will rename, delete, or add an interface as necessary to make sure
    only the primary interface exists, e.g. "wlan0" for a wireless device, e.g.
    phy0.
    """
    primaryExists = False
    renameOrRemove = list()

    for ifname in pdos.listdir(SYS_DIR):
        if ifname in EXCLUDE_IFACES:
            continue

        if getWirelessPhyName(ifname) == phy:
            if ifname == primary_interface:
                primaryExists = True
            else:
                renameOrRemove.append(ifname)

    for ifname in renameOrRemove:
        if primaryExists:
            cmd = ['iw', 'dev', ifname, 'del']
            subprocess.call(cmd)
        else:
            cmd = ['ip', 'link', 'set', 'dev', ifname, 'down', 'name',
                    primary_interface]
            subprocess.call(cmd)
            primaryExists = True

    if not primaryExists:
        cmd = ['iw', 'phy', phy, 'interface', 'add', primary_interface, 'type',
                'managed']
        subprocess.call(cmd)


def flushWirelessInterfaces(phy):
    """
    Remove all virtual interfaces associated with a wireless device.

    This should be used before giving a chute exclusive access to a device
    (e.g. monitor mode), so that it does not inherit unexpected interfaces.
    """
    for ifname in pdos.listdir(SYS_DIR):
        if ifname in EXCLUDE_IFACES:
            continue

        if getWirelessPhyName(ifname) == phy:
            cmd = ['iw', 'dev', ifname, 'del']
            subprocess.call(cmd)


def setConfig(chuteName, sections, filepath):
    cfgFile = uci.UCIConfig(filepath)

    # Set the name in the comment field.
    for config, options in sections:
        config['comment'] = chuteName

    oldSections = cfgFile.getChuteConfigs(chuteName)
    if not uci.chuteConfigsMatch(oldSections, sections):
        cfgFile.delConfigs(oldSections)
        cfgFile.addConfigs(sections)
        cfgFile.save(backupToken="paradrop", internalid=chuteName)
    else:
        # Save a backup of the file even though there were no changes.
        cfgFile.backup(backupToken="paradrop")


def readHostconfigWifi(wifi, networkDevices, builder):
    for dev in wifi:
        # The preferred method is to use the id field, which could contain many
        # different kinds of identifiers (MAC address, phy, interface, or
        # index), and use the resolveWirelessDevRef to produce a MAC
        # address-based name.  resolve that to a MAC address-based name.  Most
        # importantly, index-based names here mean host configurations can be
        # copied to different machines, but then resolved unambiguously to
        # devices.
        #
        # We can also a few other forms of identification from older
        # configuration files (macaddr, phy, or interface) and convert to
        # MAC-based name.

        if 'id' in dev:
            resolved = resolveWirelessDevRef(dev['id'], networkDevices)
            mac = resolved['mac']
        elif 'macaddr' in dev:
            mac = dev['macaddr']
        elif 'phy' in dev:
            mac = getPhyMACAddress(dev['phy'])
        elif 'interface' in dev:
            phy = getWirelessPhyName(dev['interface'])
            mac = getPhyMACAddress(phy)
        else:
            raise Exception("Missing name or address field in wifi device definition.")

        name = "wifi{}".format(mac.replace(":", ""))

        # We want to copy over all fields except name, interface, and phy.
        options = dev.copy()

        for key in ['id', 'interface', 'phy']:
            if key in options:
                del options[key]

        # Make sure macaddr is specified because that is the based way for
        # pdconf to identify the device.
        options['macaddr'] = mac

        # If type is missing, then add it because it is a required field.
        if 'type' not in options:
            options['type'] = 'auto'

        builder.add("wireless", "wifi-device", options, name=name)


def resolveWirelessDevRef(name, networkDevices):
    """
    Resolve a WiFi device reference (wlan0, phy0, 00:11:22:33:44:55, etc.) to
    the name of the device section as used by pdconf (wifiXXXXXXXXXXXX).

    Unambiguous naming is preferred going forward (either wifiXX or the MAC
    address), but to maintain backward compatibility, we attempt to resolve
    either wlanX or phyX to the MAC address of the device that currently uses
    that name.
    """
    for device in networkDevices['wifi']:
        # Construct a set of accepted identifiers for this device.
        identifiers = set()

        # MAC-based identifiers, e.g. wifi001122334455 or 00:11:22:33:44:55
        identifiers.add(device['name'])
        identifiers.add(device['mac'])

        # Index-based with deterministic ordering, e.g. pci-wifi-0
        identifiers.add(device['id'])

        # Ambiguous identifiers, e.g. phy0 or wlan0
        identifiers.add(device['phy'])
        if device['primary_interface'] is not None:
            identifiers.add(device['primary_interface'])

        # If the given name matches anything in the current set, return the
        # device name.
        if name in identifiers:
            return device

    raise DeviceNotFoundException("Could not resolve wireless device {}".format(name))


def readHostconfigWifiInterfaces(wifiInterfaces, networkDevices, builder):
    for iface in wifiInterfaces:
        # We handle nonstandard modes (e.g. Airshark) separately rather than
        # through the UCI system.
        if iface.get('mode', None) in WIFI_NONSTANDARD_MODES:
            continue

        options = iface.copy()

        # There are various ways the host configuration file may have specified
        # the WiFi device (wlan0, phy0, pci-wifi-0, 00:11:22:33:44:55, etc.).
        # Try to resolve that to a device name that pdconf will recognize.
        try:
            device = resolveWirelessDevRef(options['device'], networkDevices)
            options['device'] = device['name']
        except:
            pass

        builder.add("wireless", "wifi-iface", options)


def handleMissingWiFi(hostConfig):
    """
    Take appropriate action in response to missing WiFi devices.

    Depending on the host configuration, we may either emit a warning or reboot
    the system.
    """
    # Missing WiFi devices - check what we should do.
    action = datastruct.getValue(hostConfig, "system.onMissingWiFi")
    if action == "reboot":
        out.warn("Missing WiFi devices, system will be rebooted.")
        cmd = ["shutdown", "-r", "now"]
        subprocess.call(cmd)

    elif action == "warn":
        out.warn("Missing WiFi devices.")


def checkSystemDevices(update):
    """
    Check whether expected devices are present.

    This may reboot the machine if devices are missing and the host config is
    set to do that.
    """
    devices = update.cache_get('networkDevices')
    hostConfig = update.cache_get('hostConfig')

    if len(devices['wifi']) == 0:
        handleMissingWiFi(hostConfig)


def readHostconfigVlan(vlanInterfaces, builder):
    for interface in vlanInterfaces:
        name = interface['name']

        options = {
            'proto': interface['proto']
        }

        if interface['proto'] == 'static':
            options['ipaddr'] = interface['ipaddr']
            options['netmask'] = interface['netmask']

        # TODO: Support VLANs on interfaces other than the lan bridge.
        ifname = "br-lan.{}".format(interface['id'])
        options['ifname'] = [ifname]

        builder.add("network", "interface", options, name=name)

        if 'dhcp' in interface:
            dhcp = interface['dhcp']

            options = {}
            options['interface'] = [name]
            builder.add("dhcp", "dnsmasq", options)

            options = {
                'interface': name,
                'start': dhcp['start'],
                'limit': dhcp['limit'],
                'leasetime': dhcp['leasetime']
            }
            builder.add("dhcp", "dhcp", options, name=name)

            # Allow DNS requests.
            options = {
                'src': name,
                'proto': 'udp',
                'dest_port': 53,
                'target': 'ACCEPT'
            }
            builder.add("firewall", "rule", options)

            # Allow DHCP requests.
            options = {
                'src': name,
                'proto': 'udp',
                'dest_port': 67,
                'target': 'ACCEPT'
            }
            builder.add("firewall", "rule", options)

        # Make a zone entry with defaults.
        options = datastruct.getValue(interface,
                "firewall.defaults", {}).copy()
        options['name'] = name
        options['network'] = [name]
        builder.add("firewall", "zone", options)

        # Add forwarding entries.
        rules = datastruct.getValue(interface, "firewall.forwarding", [])
        for rule in rules:
            builder.add("firewall", "forwarding", rule)

        rules = datastruct.getValue(interface, "firewall.rules", [])
        for rule in rules:
            builder.add("firewall", "rule", rule)


class UCIBuilder(object):
    """
    UCIBuilder helps aggregate UCI configuration sections for writing to files.
    """
    FILES = ["dhcp", "network", "firewall", "wireless", "qos"]

    def __init__(self):
        self.contents = dict((f, []) for f in UCIBuilder.FILES)

    def add(self, file_, type_, options, name=None):
        """
        Add a new configuration section.
        """
        config = {"type": type_}
        if name is not None:
            config['name'] = name

        self.contents[file_].append((config, options))

    def getSections(self, file_):
        """
        Get sections associated with a single file.

        Returns: list of tuples, [(config, options)]
        """
        return self.contents[file_]

    def write(self):
        """
        Write all of the configuration sections to files.
        """
        for f in UCIBuilder.FILES:
            setConfig(constants.RESERVED_CHUTE_NAME, self.contents[f],
                    uci.getSystemPath(f))


def select_brlan_address(hostConfig):
    """
    Select IP address and netmask to use for LAN bridge.

    Behavior depends on the proto field, which can either be 'auto' or
    'static'. When proto is set to 'auto', we check the WAN interface address
    and choose either 10.0.0.0 or 192.168.0.1 to avoid conflict. Otherwise,
    when proto is set to 'static', we use the specified address.
    """
    proto = datastruct.getValue(hostConfig, 'lan.proto', 'auto')
    netmask = datastruct.getValue(hostConfig, 'lan.netmask', '255.255.255.0')
    wan_ifname = datastruct.getValue(hostConfig, 'wan.interface', 'eth0')

    if proto == 'auto':
        addresses = netifaces.ifaddresses(wan_ifname)
        ipv4_addrs = addresses.get(netifaces.AF_INET, [])

        if any(x['addr'].startswith("10.") for x in ipv4_addrs):
            return "192.168.0.1", netmask
        else:
            return "10.0.0.1", netmask

    else:
        return hostConfig['lan']['ipaddr'], netmask


#
# Chute update functions
#


def getSystemDevices(update):
    """
    Detect devices on the system.

    Store device information in cache key "networkDevices" as well as
    "networkDevicesByName".
    """
    devices = detectSystemDevices()

    devicesByName = dict()
    for dtype, dlist in six.iteritems(devices):
        for dev in dlist:
            name = dev['name']
            if name in devicesByName:
                out.warn("Multiple network devices named {}".format(name))
            devicesByName[name] = dev

    update.cache_set('networkDevices', devices)
    update.cache_set('networkDevicesByName', devicesByName)


def setSystemDevices(update):
    """
    Initialize system configuration files.

    This section should only be run for host configuration updates.

    Creates basic sections that all chutes require such as the "wan" interface.
    """
    hostConfig = update.cache_get('hostConfig')
    networkDevices = update.cache_get('networkDevices')

    builder = UCIBuilder()

    # This section defines the default input, output, and forward policies for
    # the firewall.
    options = datastruct.getValue(hostConfig, "firewall.defaults", {})
    builder.add("firewall", "defaults", options)

    def zoneFirewallSettings(name):
        # Create zone entry with defaults (input, output, forward policies and
        # other configuration).
        #
        # Make a copy of the object from hostconfig because we modify it.
        options = datastruct.getValue(hostConfig,
                name+".firewall.defaults", {}).copy()
        options['name'] = name
        options['network'] = [name]
        builder.add("firewall", "zone", options)

        # Add forwarding entries (rules that allow traffic to move from one
        # zone to another).
        rules = datastruct.getValue(hostConfig, name+".firewall.forwarding", [])
        for rule in rules:
            builder.add("firewall", "forwarding", rule)

    if 'wan' in hostConfig:
        options = dict()
        options['ifname'] = hostConfig['wan']['interface']
        options['proto'] = "dhcp"
        builder.add("network", "interface", options, name="wan")

        zoneFirewallSettings("wan")

        options = {
            "enabled": 0
        }
        builder.add("qos", "interface", options, name="wan")

    if 'lan' in hostConfig:
        options = dict()
        options['type'] = "bridge"
        options['bridge_empty'] = "1"
        options['proto'] = 'static'
        options['ipaddr'], options['netmask'] = select_brlan_address(hostConfig)
        options['ifname'] = hostConfig['lan']['interfaces']
        builder.add("network", "interface", options, name="lan")

        if 'dhcp' in hostConfig['lan']:
            dhcp = hostConfig['lan']['dhcp']

            options = {
                'interface': 'lan',
                'domain': settings.LOCAL_DOMAIN
            }
            builder.add("dhcp", "dnsmasq", options)

            options = {
                'interface': 'lan',
                'start': dhcp['start'],
                'limit': dhcp['limit'],
                'leasetime': dhcp['leasetime']
            }
            builder.add("dhcp", "dhcp", options, name="lan")

            options = {
                'name': settings.LOCAL_DOMAIN,
                'ip': hostConfig['lan']['ipaddr']
            }
            builder.add("dhcp", "domain", options)

        zoneFirewallSettings("lan")

        options = {
            "enabled": 0
        }
        builder.add("qos", "interface", options, name="lan")

    # Automatically generate loopback section.  There is generally not much to
    # configure for loopback, but we could add support to the host
    # configuration.
    options = {
        'ifname': ['lo'],
        'proto': 'static',
        'ipaddr': '127.0.0.1',
        'netmask': '255.0.0.0'
    }
    builder.add("network", "interface", options, name="loopback")

    options = {
        'name': 'loopback',
        'masq': '0',
        'conntrack': '1',
        'input': 'ACCEPT',
        'forward': 'ACCEPT',
        'output': 'ACCEPT',
        'network': ['loopback']
    }
    builder.add("firewall", "zone", options)

    wifi = hostConfig.get('wifi', [])
    try:
        readHostconfigWifi(wifi, networkDevices, builder)
    except DeviceNotFoundException:
        handleMissingWiFi(hostConfig)

    wifiInterfaces = hostConfig.get('wifi-interfaces', [])
    readHostconfigWifiInterfaces(wifiInterfaces, networkDevices, builder)

    vlanInterfaces = hostConfig.get('vlan-interfaces', [])
    readHostconfigVlan(vlanInterfaces, builder)

    # Add additional firewall rules.
    rules = datastruct.getValue(hostConfig, "firewall.rules", [])
    for rule in rules:
        builder.add("firewall", "rule", rule)

    # Write all of the changes to UCI files at once.
    builder.write()


def get_hardware_serial():
    """
    Get hardware serial number.

    The most reliable way we have that works across many hardware platforms is
    to check the eth0 MAC address.

    Returns a numeric serial number.
    """
    addr = getMACAddress("eth0")
    if addr is None:
        return 0
    else:
        return int(addr.translate(None, ":.- "), 16)


def get_machine_id():
    """
    Return unique machine identifier.

    This is software-based but fairly standardized from the /etc/machine-id file.
    We can potentially rely on this for uniquely identifying a node.
    """
    if os.path.isfile("/etc/machine-id"):
        return readSysFile("/etc/machine-id")
    elif os.path.isfile("/var/lib/dbus/machine-id"):
        return readSysFile("/var/lib/dbus/machine-id")
    else:
        return None



'''
Stateful, singleton, paradrop daemon command center.
See docstring for NexusBase class for information on settings.

SETTINGS QUICK REFERENCE:
    # assuming the following import
    from paradrop.base import nexus

    nexus.core.info.version
    nexus.core.info.pdid
'''

import os
import yaml
import json

import smokesignal
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, returnValue

from . import output, settings
from paradrop.lib.utils import pdosq

# Global access. Assign this wherever you instantiate the Nexus object:
#       nexus.core = MyNexusSubclass()
core = None


class NexusBase(object):
    '''
    Resolving these values to their final forms:
        1 - module imported, initial values assigned(as written below)
        2 - class is instatiated, passed settings to replace values
        3 - instance chooses appropriate values based on current state(production or local)
            Each category has its own method for initialization here
            (see: resolveNetwork, resolvePaths)

    '''
    VERSION = 1                                         # nexus.core.info.version
    PDID = None                                         # nexus.core.info.pdid


    def __init__(self, stealStdio=True, printToConsole=True):
        '''
        The one big thing this function leaves out is reactor.start(). Call this externally
        *after * initializing a nexus object.
        '''
        self.session = None
        self.wamp_connected = False
        self.jwt_valid = False

        self.info = AttrWrapper()
        resolveInfo(self, settings.CONFIG_FILE)
        self.info.setOnChange(self.onInfoChange)

        # initialize output. If filepath is set, logs to file.
        # If stealStdio is set intercepts all stderr and stdout and interprets it internally
        # If printToConsole is set (defaults True) all final output is rendered to stdout
        output.out.startLogging(filePath=settings.LOG_DIR, stealStdio=stealStdio, printToConsole=printToConsole)

        # register onStop for the shutdown call
        reactor.addSystemEventTrigger('before', 'shutdown', self.onStop)

        # The reactor needs to be runnnig before this call is fired, since we start the session
        # here. Assuming callLater doesn't fire until thats happened
        reactor.callLater(0, self.onStart)


    def onStart(self):
        pdid = self.info.pdid if self.provisioned() else 'UNPROVISIONED'
        output.out.usage('%s coming up' % (pdid))


    def onStop(self):
        self.save()

        output.out.usage('%s going down' % (self.info.pdid))
        smokesignal.clear_all()
        output.out.endLogging()


    @inlineCallbacks
    def connect(self, sessionClass, debug=False):
        '''
        Takes the given session class and attempts to connect to the crossbar fabric.

        If an existing session is connected, it is cleanly closed.
        '''

        if (self.session is not None):
            yield self.session.leave()

        self.wamp_connected = False

        output.out.info('Connecting to wamp router at URI: %s' % str(self.info.wampRouter))

        # Setting self.session here only works for the first connection but
        # becomes stale if the connection fails and we reconnect.
        # In that case a new session object is automatically created.
        # For this reason, we also update this session reference in BaseSession.onJoin.
        self.session = yield sessionClass.start(self.info.wampRouter, self.info.pdid, debug=debug)
        returnValue(self.session)


    def onInfoChange(self, key, value):
        '''
        Called when an internal setting is changed. Trigger a save automatically.
        '''
        self.save()


    def save(self):
        ''' Ehh. Ideally this should happen asynchronously. '''
        saveDict = self.info.__dict__['contents']
        saveDict['version'] = self.info.version

        writeYaml(saveDict, settings.CONFIG_FILE)

    #########################################################
    # High Level Methods
    #########################################################

    def provisioned(self):
        '''
        Checks if this[whatever] appears to be provisioned or not
        '''

        return self.info.pdid is not None

    def provision(self, pdid, pdserver=settings.PDSERVER, wampRouter=settings.WAMP_ROUTER):
        self.info.pdid = pdid
        self.info.pdserver = pdserver
        self.info.wampRouter = wampRouter

    #########################################################
    # Keys
    #########################################################

    def saveKey(self, key, name):
        ''' Save the key with the given name. Overwrites by default '''
        path = settings.KEY_DIR + name

        with open(path, 'wb') as f:
            f.write(key)

    def getKey(self, name):
        ''' Returns the given key or None '''
        path = settings.KEY_DIR + name

        if os.path.isfile(path):
            with open(path, 'rb') as f:
                return f.read()

        return None

    #########################################################
    # Misc
    #########################################################

    def __repr__(self):
        ''' Dump everything '''
        dic = dict(info=self.info.contents)
        return json.dumps(dic, sort_keys=True, indent=4)

#########################################################
# Utils
#########################################################

class AttrWrapper(object):
    '''
    Simple attr interceptor to make accessing settings simple.

    Stores values in an internal dict called contents.

    Does not allow modification once _lock() is called. Respect it.

    Once you've filled it up with the appropriate initial values, set
    onChange to assign
    '''

    def __init__(self):
        self.__dict__['contents'] = {}

        # Called when a value changes unless None
        self.__dict__['onChange'] = None

        # Lock the contents of this wrapper. Can read valued, cant write them
        self.__dict__['locked'] = False

    def _lock(self):
        self.__dict__['locked'] = True

    def setOnChange(self, func):
        assert(callable(func))
        self.__dict__['onChange'] = func

    def __repr__(self):
        return str(self.contents)

    def __getattr__(self, name):
        return self.__dict__['contents'][name]

    def __setattr__(self, k, v):
        if self.__dict__['locked']:
            raise AttributeError('This attribute wrapper is locked. You cannot change its values.')

        self.contents[k] = v

        if self.__dict__['onChange'] is not None:
            self.__dict__['onChange'](k, v)


def resolveInfo(nexus, path):
    '''
    Given a path to the config file, load its contents and assign it to the
    config file as appropriate.
    '''

    # Check to make sure we have a default settings file
    if not os.path.isfile(path):
        createDefaultInfo(path)

    contents = pdosq.read_yaml_file(path)

    # Sanity check contents of info and throw it out if bad
    if not validateInfo(contents):
        output.out.err('Saved configuration data invalid, destroying it.')
        os.remove(path)
        createDefaultInfo(path)
        contents = pdosq.read_yaml_file(path, default={})
        writeYaml(contents, path)

    nexus.info.pdid = contents['pdid']
    nexus.info.version = contents['version']
    nexus.info.pdserver = contents['pdserver']
    nexus.info.wampRouter = contents['wampRouter']


def createDefaultInfo(path):
    default = {
        'version': 1,
        'pdid': None,
        'pdserver': settings.PDSERVER,
        'wampRouter': settings.WAMP_ROUTER
    }

    writeYaml(default, path)


def validateInfo(contents):
    '''
    Error checking on the read YAML file. This is a temporary method.

    :
        param contents:
            the read - in yaml to check
    :
        type contents:
            dict.
    :
        returns:
            True if valid, else false
    '''
    INFO_REQUIRES = ['version', 'pdid', 'pdserver', 'wampRouter']

    for k in INFO_REQUIRES:
        if k not in contents:
            output.out.err('Contents is missing: ' + str(k))
            return False

    return True


def writeYaml(contents, path):
    ''' Overwrites content with YAML representation at given path '''
    # print 'Writing ' + str(contents) + ' to path ' + str(path)

    with open(path, 'w') as f:
        f.write(yaml.safe_dump(contents, default_flow_style=False))



"""
    dockerconfig module:
        This module contains all of the knowledge of how to take internal pdfcd
        representation of configurations of chutes and translate them into specifically
        what docker needs to function properly, whether that be in the form of dockerfiles
        or the HostConfig JSON object known at init time of the chute.
"""

import os
import random
from io import BytesIO

from paradrop.base.output import out
from paradrop.base import settings
from paradrop.core.container import dockerapi
from paradrop.lib.utils import pdos, pdosq


def generateToken(bits=128):
    return "{:x}".format(random.getrandbits(bits))


def getVirtPreamble(update):
    """
    Prepare various settings for Docker containers.
    """
    extDataDir = settings.EXTERNAL_DATA_DIR.format(chute=update.new.name)
    extDataDir = os.path.expanduser(extDataDir)
    intDataDir = getattr(update.new, 'dataDir', settings.INTERNAL_DATA_DIR)

    extSystemDir = settings.EXTERNAL_SYSTEM_DIR.format(chute=update.new.name)
    extSystemDir = os.path.expanduser(extSystemDir)
    intSystemDir = getattr(update.new, 'systemDir', settings.INTERNAL_SYSTEM_DIR)

    volumes = {
        extDataDir: {
            'bind': intDataDir,
            'mode': 'rw'
        },

        extSystemDir: {
            'bind': intSystemDir,
            'mode': 'ro'
        }
    }

    # Prepare port bindings while there might still be an old version of the
    # chute running so that we can inherit any dynamic port settings.
    for service in update.new.get_services():
        bindings = dockerapi.prepare_port_bindings(service)
        update.cache_set("portBindings:{}".format(service.name), bindings)

    update.cache_set('volumes', volumes)
    update.cache_set('internalDataDir', intDataDir)
    update.cache_set('externalDataDir', extDataDir)
    update.cache_set('internalSystemDir', intSystemDir)
    update.cache_set('externalSystemDir', extSystemDir)

    # Reuse previous token if it exists. This is not ideal but works for
    # updates that do not recreate the container with updated environment
    # variables. Stop then start is one such sequence where environment
    # variables are not updated.
    token = None
    if update.old is not None:
        token = update.old.getCache('apiToken')
    if token is None:
        token = generateToken()
    update.cache_set('apiToken', token)

    # Deprecated: we set these values in the chute cache here because there is
    # still code that depends on reading this from the chute storage.  This can
    # be removed when the corresponding call to getCache is removed.
    update.new.setCache('externalSystemDir', extSystemDir)
    update.new.setCache('apiToken', token)

    if not hasattr(update, 'dockerfile'):
        return
    if update.dockerfile is None:
        return
    else:
        out.info('Using prexisting dockerfile.\n')
        update.dockerfile = BytesIO(update.dockerfile.encode('utf-8'))


def createVolumeDirs(update):
    """
    Create directories required by the chute.
    """
    extDataDir = update.cache_get('externalDataDir')
    extSystemDir = update.cache_get('externalSystemDir')

    if update.updateType == 'delete':
        pdos.remove(extDataDir, suppressNotFound=True)
        pdos.remove(extSystemDir, suppressNotFound=True)
    else:
        pdosq.makedirs(extDataDir)
        pdosq.makedirs(extSystemDir)
        try:
            os.chown(extDataDir, settings.CONTAINER_UID, -1)
        except:
            # chown is not permitted in strict confinement.
            # TODO: do we need to chmod the data directory, or does it work
            # fine regardless?
            pass


def abortCreateVolumeDirs(update):
    # If there is no old version of the chute, then clean up directories that
    # we created.  Otherwise, leave them in place for the old version.
    if update.old is None:
        extDataDir = update.cache_get('externalDataDir')
        pdos.remove(extDataDir)

        extSystemDir = update.cache_get('externalSystemDir')
        pdos.remove(extSystemDir)



"""
This module exposes device configuration.

Endpoints for these functions can be found under /api/v1/config.
"""

import json
from klein import Klein
from twisted.internet import reactor
from twisted.internet.defer import DeferredList, inlineCallbacks, returnValue

from paradrop.base import constants, nexus, settings
from paradrop.base.output import out
from paradrop.base.pdutils import timeint
from paradrop.core.config import hostconfig
from paradrop.core.agent.http import PDServerRequest
from paradrop.core.agent.provisioning import read_provisioning_result
from paradrop.core.agent.reporting import sendNodeIdentity, sendStateReport
from paradrop.core.agent.wamp_session import WampSession
from paradrop.confd import client as pdconf_client
from paradrop.lib.misc import ssh_keys
from paradrop.lib.misc.governor import GovernorClient

from . import cors


class ConfigApi(object):
    """
    Configuration API.

    This class handles HTTP API calls related to router configuration.
    """

    routes = Klein()

    def __init__(self, update_manager, update_fetcher):
        self.update_manager = update_manager
        self.update_fetcher = update_fetcher

    @routes.route('/hostconfig', methods=['PUT'])
    def update_hostconfig(self, request):
        """
        Replace the device's host configuration.

        **Example request**:

        .. sourcecode:: http

           PUT /api/v1/config/hostconfig
           Content-Type: application/json

           {
             "firewall": {
               "defaults": {
                 "forward": "ACCEPT",
                 "input": "ACCEPT",
                 "output": "ACCEPT"
               }
             },
             ...
           }

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             change_id: 1
           }

        For a complete example, please see the Host Configuration section.
        """
        cors.config_cors(request)
        body = json.loads(request.content.read().decode('utf-8'))
        config = body['config']

        update = dict(updateClass='ROUTER',
                      updateType='sethostconfig',
                      name=constants.RESERVED_CHUTE_NAME,
                      tok=timeint(),
                      hostconfig=config)

        # We will return the change ID to the caller for tracking and log
        # retrieval.
        update['change_id'] = self.update_manager.assign_change_id()

        self.update_manager.add_update(**update)

        result = {
            'change_id': update['change_id']
        }
        request.setHeader('Content-Type', 'application/json')
        return json.dumps(result)

    @routes.route('/hostconfig', methods=['GET'])
    def get_hostconfig(self, request):
        """
        Get the device's current host configuration.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/config/hostconfig

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "firewall": {
               "defaults": {
                 "forward": "ACCEPT",
                 "input": "ACCEPT",
                 "output": "ACCEPT"
               }
             },
             ...
           }

        For a complete example, please see the Host Configuration section.
        """
        cors.config_cors(request)
        config = hostconfig.prepareHostConfig()
        request.setHeader('Content-Type', 'application/json')
        return json.dumps(config, separators=(',',':'))

    @routes.route('/new-config', methods=['GET'])
    def new_config(self, request):
        """
        Generate a new node configuration based on the hardware.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/config/new_config

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "firewall": {
               "defaults": {
                 "forward": "ACCEPT",
                 "input": "ACCEPT",
                 "output": "ACCEPT"
               }
             },
             ...
           }

        For a complete example, please see the Host Configuration section.
        """
        cors.config_cors(request)
        config = hostconfig.prepareHostConfig(hostConfigPath='/dev/null')
        request.setHeader('Content-Type', 'application/json')
        return json.dumps(config, separators=(',',':'))

    @routes.route('/pdid', methods=['GET'])
    def get_pdid(self, request):
        """
        Get the device's current ParaDrop ID. This is the identifier assigned
        by the cloud controller.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/config/pdid

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             pdid: "5890e1e5ab7e317e6c6e049f"
           }
        """
        cors.config_cors(request)
        pdid = nexus.core.info.pdid
        if pdid is None:
            pdid = ""
        request.setHeader('Content-Type', 'application/json')
        return json.dumps({'pdid': pdid})


    @routes.route('/provision', methods=['POST'])
    def provision(self, request):
        """
        Provision the device with credentials from a cloud controller.
        """
        cors.config_cors(request)
        body = json.loads(request.content.read().decode('utf-8'))
        routerId = body['routerId']
        apitoken = body['apitoken']
        pdserver = body['pdserver']
        wampRouter = body['wampRouter']

        changed = False
        if routerId != nexus.core.info.pdid \
            or pdserver != nexus.core.info.pdserver \
            or wampRouter != nexus.core.info.wampRouter:
            if pdserver and wampRouter:
                nexus.core.provision(routerId, pdserver, wampRouter)
            else:
                nexus.core.provision(routerId)
            changed = True

        if apitoken != nexus.core.getKey('apitoken'):
            nexus.core.saveKey(apitoken, 'apitoken')
            changed = True

        if changed:
            PDServerRequest.resetToken()
            nexus.core.jwt_valid = False

            def set_update_fetcher(session):
                session.set_update_fetcher(self.update_fetcher)

            @inlineCallbacks
            def start_polling(result):
                yield self.update_fetcher.start_polling()

            def send_response(result):
                response = dict()
                response['provisioned'] = True
                response['httpConnected'] = nexus.core.jwt_valid
                response['wampConnected'] = nexus.core.wamp_connected
                request.setHeader('Content-Type', 'application/json')
                return json.dumps(response)

            wampDeferred = nexus.core.connect(WampSession)
            wampDeferred.addCallback(set_update_fetcher)

            httpDeferred = sendStateReport()
            httpDeferred.addCallback(start_polling)

            identDeferred = sendNodeIdentity()

            dl = DeferredList([wampDeferred, httpDeferred, identDeferred],
                    consumeErrors=True)
            dl.addBoth(send_response)
            reactor.callLater(6, dl.cancel)
            return dl
        else:
            return json.dumps({'success': False,
                               'message': 'No change on the provision parameters'})

    @routes.route('/provision', methods=['GET'])
    def get_provision(self, request):
        """
        Get the provision status of the device.
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        result = read_provisioning_result()
        result['routerId'] = nexus.core.info.pdid
        result['pdserver'] = nexus.core.info.pdserver
        result['wampRouter'] = nexus.core.info.wampRouter
        apitoken = nexus.core.getKey('apitoken')
        result['provisioned'] = (result['routerId'] is not None and \
                                 apitoken is not None)
        result['httpConnected'] = nexus.core.jwt_valid
        result['wampConnected'] = nexus.core.wamp_connected
        return json.dumps(result)

    @routes.route('/settings', methods=['GET'])
    def get_settings(self, request):
        """
        Get current values of system settings.

        These are the values from paradrop.base.settings. Settings are loaded
        at system initialization from the settings.ini file and environment
        variables. They are intended to be read-only after initialization.

        This endpoint returns the settings as a dictionary with lowercase field
        names.

        Example:
        {
            "portal_server_port": 8080,
            ...
        }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        result = {}
        for name, value in settings.iterate_module_attributes(settings):
            result[name.lower()] = value

        return json.dumps(result)

    @routes.route('/startUpdate', methods=['POST'])
    def start_update(self, request):
        cors.config_cors(request)
        self.update_manager.startUpdate()
        request.setHeader('Content-Type', 'application/json')
        return json.dumps({'success': True})


    @routes.route('/factoryReset', methods=['POST'])
    @inlineCallbacks
    def factory_reset(self, request):
        """
        Initiate the factory reset process.
        """
        cors.config_cors(request)
        update = dict(updateClass='ROUTER',
                      updateType='factoryreset',
                      name='PARADROP',
                      tok=timeint())
        update = yield self.update_manager.add_update(**update)
        returnValue(json.dumps(update.result))

    @routes.route('/pdconf', methods=['GET'])
    def pdconf(self, request):
        """
        Get configuration sections from pdconf.

        This returns a list of configuration sections and whether they were
        successfully applied. This is intended for debugging purposes.
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')
        return pdconf_client.systemStatus()

    @routes.route('/pdconf', methods=['PUT'])
    def pdconf_reload(self, request):
        """
        Trigger pdconf to reload UCI configuration files.

        Trigger pdconf to reload UCI configuration files and return the status.
        This function is intended for low-level debugging of the paradrop
        pdconf module.
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')
        return pdconf_client.reloadAll()

    @routes.route('/sshKeys/<user>', methods=['GET', 'POST'])
    def sshKeys(self, request, user):
        """
        Manage list of authorized keys for SSH access.
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        if request.method == "GET":
            try:
                if GovernorClient.isAvailable():
                    client = GovernorClient()
                    keys = client.listAuthorizedKeys(user)

                    # Move keys to the root because the API consumer is
                    # expecting a list, not an object.
                    if 'keys' in keys:
                        keys = keys['keys']
                else:
                    keys = ssh_keys.getAuthorizedKeys(user)
                return json.dumps(keys)
            except Exception as e:
                out.warn(str(e))
                request.setResponseCode(404)
                return json.dumps({'message': str(e)})
        else:
            body = json.loads(request.content.read().decode('utf-8'))
            key = body['key'].strip()

            try:
                if GovernorClient.isAvailable():
                    client = GovernorClient()
                    client.addAuthorizedKey(key, user)
                else:
                    ssh_keys.addAuthorizedKey(key, user)
                return json.dumps(body)
            except Exception as e:
                out.warn(str(e))
                request.setResponseCode(404)
                return json.dumps({'message': str(e)})



'''
Wamp utility methods.
'''

from autobahn.twisted import wamp, websocket
from autobahn.twisted.wamp import ApplicationSession
from autobahn.wamp.types import ComponentConfig
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, Deferred
from twisted.internet.protocol import ReconnectingClientFactory
from twisted.internet.ssl import ClientContextFactory

from . import nexus
from .output import out


# pylint: disable=inconsistent-mro
class BaseClientFactory(websocket.WampWebSocketClientFactory, ReconnectingClientFactory):
    # factor and jitter are two variables that control the exponential backoff.
    # The default values in ReconnectingClientFactory seem reasonable.
    initialDelay = 1
    maxDelay = 60

    def clientConnectionFailed(self, connector, reason):
        out.info("Connection failed with reason: " + str(reason))
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)

    def clientConnectionLost(self, connector, reason):
        out.info("Connection lost with reason: " + str(reason))
        ReconnectingClientFactory.clientConnectionLost(self, connector, reason)


class BaseSessionFactory(wamp.ApplicationSessionFactory):
    def __init__(self, config, deferred=None):
        super(BaseSessionFactory, self).__init__(config)
        self.dee = deferred

    def __call__(self, *args, **kwargs):
        sess = super(BaseSessionFactory, self).__call__(*args, **kwargs)
        sess.dee = self.dee
        return sess


class BaseSession(ApplicationSession):

    """ Temporary base class for crossbar implementation """

    def __init__(self, config=None):
        ApplicationSession.__init__(self, config=config)
        self.pdid = config.extra

    @classmethod
    def start(klass, address, pdid, realm='paradrop', start_reactor=False,
              debug=False, extra=None, reconnect=True):
        '''
        Creates a new instance of this session and attaches it to the router
        at the given address and realm.

        reconnect: The session will attempt to reconnect on connection failure
            and continue trying indefinitely.
        '''
        # Enable log messages of autobahn for debugging
        #import txaio
        #txaio.start_logging()

        dee = Deferred()

        component_config = ComponentConfig(realm=u''+realm, extra=u''+pdid)
        session_factory = BaseSessionFactory(config=component_config, deferred=dee)
        session_factory.session = klass

        transport_factory = BaseClientFactory(session_factory, url=address)
        if not reconnect:
            transport_factory.maxRetries = 0
        transport_factory.setProtocolOptions(autoPingInterval=8., autoPingTimeout=4.,)
        context_factory = ClientContextFactory()
        websocket.connectWS(transport_factory, context_factory)

        if start_reactor:
            reactor.run()

        return dee

        # This the the recommended way to start the WAMP component,
        # but it is friendly to customize the component
        #runner = ApplicationRunner(url=u''+address, realm=u''+realm)
        #return runner.run(klass, start_reactor=start_reactor, auto_reconnect=reconnect)

    def leave(self):
        # Do not retry if explicitly asked to leave.
        self._transport.factory.maxRetries = 0
        nexus.core.session = None
        nexus.core.wamp_connected = False
        super(BaseSession, self).leave()

    @inlineCallbacks
    def onJoin(self, details):
        out.info(str(self.__class__.__name__) + ' crossbar session connected')

        # Update global session reference.  It's hacky, but we do the import
        # here to solve the circular import problem.  TODO Refactor.
        nexus.core.session = self
        nexus.core.wamp_connected = True

        # Reset exponential backoff timer after a successful connection.
        self._transport.factory.resetDelay()

        # Inform whoever created us that the session has finished connecting.
        # Useful in situations where you need to fire off a single call and not a
        # full wamplet
        if self.dee is not None:
            yield self.dee.callback(self)

    ###################################################
    # Overridden CX interaction methods
    ###################################################

    '''
    Note: this first set of methods have all the REAL implemnetion of caller identification:
    the caller's information is always passed along for every call. In the crossbar way of doing
    things, however, whats passed along is a session id and not our pdid.

    The second set of methods is temporary in that it manually passes the
    current sessions' pdid. This is not secure, but will have to do for now in the
    absence of crossbar router changes.
    '''

    def publish(self, topic, *args, **kwargs):
        # kwargs['options'] = PublishOptions(disclose_me=True)
        args = (self.pdid,) + args
        topic = _prepend(self.pdid, topic)
        # out.info('cxbr: (%s) publish (%s)' % (self.pdid, topic,))
        return ApplicationSession.publish(self, topic, *args, **kwargs)

    def subscribe(self, handler, topic=None, options=None):
        topic = _prepend(self.pdid, topic)
        out.info('cxbr: (%s) subscribe (%s)' % (self.pdid, topic,))
        return ApplicationSession.subscribe(self, handler, topic=topic, options=options)

    def call(self, procedure, *args, **kwargs):
        # kwargs['options'] = CallOptions(disclose_me=True)
        args = (self.pdid,) + args
        procedure = _prepend(self.pdid, procedure)
        # out.info('cxbr: (%s) calling (%s)' % (self.pdid, procedure,))
        return ApplicationSession.call(self, procedure, *args, **kwargs)

    def register(self, endpoint, procedure=None, options=None):
        # options = RegisterOptions(details_arg='session')
        procedure = _prepend(self.pdid, procedure)
        out.info('cxbr: (%s) register (%s)' % (self.pdid, procedure,))
        return ApplicationSession.register(self, endpoint, procedure=procedure, options=options)

    ###################################################
    # Access to the original methods, without convenience modifiers
    ###################################################
    def stockPublish(self, topic, *args, **kwargs):
        return ApplicationSession.publish(self, u''+topic, *args, **kwargs)

    def stockSubscribe(self, handler, topic=None, options=None):
        return ApplicationSession.subscribe(self, handler, topic=u''+topic, options=options)

    def stockCall(self, procedure, *args, **kwargs):
        return ApplicationSession.call(self, u''+procedure, *args, **kwargs)

    def stockRegister(self, endpoint, procedure=None, options=None):
        return ApplicationSession.register(self, endpoint, procedure=u''+procedure, options=options)


def _prepend(pdid, topic):
    '''
    In order to make subscription and execution code cleaner, this method automatically
    injects this classes pdid to the end of any publish or register call.

    The topic is also converted to a unicode string.
    No consideration is given to 'valid' topics-- thats on you.
    '''
    return u'' + topic + '.' + pdid



import re

from .base import ConfigObject, ConfigOption
from .command import Command


class ConfigInterface(ConfigObject):
    typename = "interface"

    # Don't let qos:interface override this class.
    maskable = False

    options = [
        ConfigOption(name="proto", required=True),
        ConfigOption(name="ifname", type=list, default=[]),
        ConfigOption(name="type"),
        ConfigOption(name="bridge_empty", type=bool, default=False),
        ConfigOption(name="enabled", type=bool, default=True),

        # Options for "static" protocol:
        ConfigOption(name="ipaddr"),
        ConfigOption(name="netmask"),
        ConfigOption(name="gateway")
    ]

    # Match interface names that include a VLAD ID, e.g. eth0.10
    DEV_PLUS_VID = re.compile(r"([a-z0-9\-]+)\.(\d+)")

    def setup(self):
        # For which interface are we setting the IP address?  If it's a bridge,
        # we create a bridge interface and set its address; otherwise, we take
        # the "ifname" option.
        if self.type == "bridge":
            self.config_ifname = "br-{}".format(self.name)
        else:
            if len(self.ifname) == 0:
                raise Exception("missing ifname in interface {}".format(self.name))
            self.config_ifname = self.ifname[0]

        # Set to True if we created a new interface.
        self._created = False

    def addToBridge(self, ifname):
        """ Generate commands to add ifname to bridge. """
        commands = list()

        cmd = ["ip", "link", "set", "dev", ifname, "promisc", "on"]
        commands.append((self.PRIO_CONFIG_IFACE, Command(cmd, self)))

        cmd = ["ip", "link", "set", "dev", ifname, "up"]
        commands.append((self.PRIO_CONFIG_IFACE, Command(cmd, self)))

        cmd = ["ip", "link", "set", "dev", ifname, "master",
                self.config_ifname]
        commands.append((self.PRIO_CONFIG_IFACE, Command(cmd, self)))

        return commands

    def removeFromBridge(self, ifname):
        """ Generate commands to add ifname to bridge. """
        commands = list()

        cmd = ["ip", "link", "set", "dev", ifname, "promisc", "off"]
        commands.append((-self.PRIO_CONFIG_IFACE, Command(cmd, self)))

        cmd = ["ip", "link", "set", "dev", ifname, "down"]
        commands.append((-self.PRIO_CONFIG_IFACE, Command(cmd, self)))

        cmd = ["ip", "link", "set", "dev", ifname, "nomaster"]
        commands.append((-self.PRIO_CONFIG_IFACE, Command(cmd, self)))

        return commands

    def apply(self, allConfigs):
        commands = list()

        if self.type == "bridge":
            # Create bridge interface if there is at least one physical
            # interface or bridge_empty is set to true.
            if len(self.ifname) > 0 or self.bridge_empty:
                cmd = ["ip", "link", "add", "name", self.config_ifname, "type",
                       "bridge"]
                commands.append((self.PRIO_CREATE_IFACE, Command(cmd, self)))

                self._created = True

            # Add all of the interfaces to the bridge.
            for ifname in self.ifname:
                commands.extend(self.addToBridge(ifname))
        else:
            # If the ifname contains a VLAN ID, e.g. eth0.5, then create a vlan
            # interface.
            match = ConfigInterface.DEV_PLUS_VID.match(self.config_ifname)
            if match is not None:
                ifname = match.group(1)
                vlan_id = match.group(2)

                cmd = ["ip", "link", "add", "link", ifname, "name",
                        self.config_ifname, "type", "vlan", "id", vlan_id]
                commands.append((self.PRIO_CREATE_VLAN, Command(cmd, self)))

                self._created = True

        if self.proto == "static":
            cmd = ["ip", "addr", "flush", "dev", self.config_ifname]
            commands.append((self.PRIO_CONFIG_IFACE, Command(cmd, self)))

            cmd = ["ip", "addr", "add",
                   "{}/{}".format(self.ipaddr, self.netmask),
                   "dev", self.config_ifname]
            commands.append((self.PRIO_CONFIG_IFACE, Command(cmd, self)))

            updown = "up" if self.enabled else "down"
            cmd = ["ip", "link", "set", "dev", self.config_ifname, updown]
            commands.append((self.PRIO_CONFIG_IFACE, Command(cmd, self)))

            if self.gateway is not None:
                cmd = ["ip", "route", "add", "default", "via", self.gateway,
                       "dev", self.config_ifname]
                commands.append((self.PRIO_CONFIG_IFACE, Command(cmd, self)))

        return commands

    def revert(self, allConfigs):
        commands = list()

        if self.proto == "static":
            if self.gateway is not None:
                cmd = ["ip", "route", "del", "default", "via", self.gateway,
                       "dev", self.config_ifname]
                commands.append((-self.PRIO_CREATE_IFACE, Command(cmd, self)))

            # Remove the IP address that we added.
            cmd = ["ip", "addr", "del",
                   "{}/{}".format(self.ipaddr, self.netmask),
                   "dev", self.config_ifname]
            commands.append((-self.PRIO_CONFIG_IFACE, Command(cmd, self)))

        if self.type == "bridge":
            # Remove all of the interfaces from the bridge.
            for ifname in self.ifname:
                commands.extend(self.removeFromBridge(ifname))

        if self._created:
            cmd = ["ip", "link", "delete", self.config_ifname]
            commands.append((-self.PRIO_CREATE_IFACE, Command(cmd, self)))

        return commands

    def updateApply(self, new, allConfigs):
        # Major changes require reverting the current config and loading the
        # new one.
        if self.proto != new.proto or self.type != new.type:
            return self.apply(allConfigs)

        commands = list()

        # May have changed the IP address.
        if self.proto == "static":
            if new.ipaddr != self.ipaddr or new.netmask != self.netmask:
                cmd = ["ip", "addr", "change",
                       "{}/{}".format(new.ipaddr, new.netmask),
                       "dev", new.config_ifname]
                commands.append((self.PRIO_CONFIG_IFACE, Command(cmd, self)))

            if new.gateway != self.gateway:
                if new.gateway is not None:
                    cmd = ["ip", "route", "add", "default", "via", new.gateway,
                            "dev", new.config_ifname]
                    commands.append((self.PRIO_CONFIG_IFACE,
                        Command(cmd, new)))

        if self.type == "bridge":
            old_ifnames = set(self.ifname)
            new_ifnames = set(new.ifname)

            # The bridge may not exist if this is the first physical interface
            # and bridge_empty is False.
            if len(self.ifname) == 0 and not self.bridge_empty:
                cmd = ["ip", "link", "add", "name", self.config_ifname, "type",
                        "bridge"]
                commands.append((self.PRIO_CREATE_IFACE, Command(cmd, self)))

                self._created = True

            # Add interfaces that were not in the old bridge.
            for ifname in (new_ifnames - old_ifnames):
                commands.extend(self.addToBridge(ifname))

        return commands

    def updateRevert(self, new, allConfigs):
        # Major changes require reverting the current config and loading the
        # new one.
        if self.proto != new.proto or self.type != new.type:
            return self.revert(allConfigs)

        commands = list()

        # May have changed the IP address.
        if self.proto == "static":
            if new.gateway != self.gateway:
                if self.gateway is not None:
                    cmd = ["ip", "route", "del", "default", "via", self.gateway,
                           "dev", self.config_ifname]
                    commands.append((-self.PRIO_CONFIG_IFACE,
                        Command(cmd, self)))

            # Delete the old IP address because "ip addr change" seems to leave
            # it in place.
            if new.ipaddr != self.ipaddr or new.netmask != self.netmask:
                cmd = ["ip", "addr", "del",
                       "{}/{}".format(self.ipaddr, self.netmask),
                       "dev", self.config_ifname]
                commands.append((self.PRIO_CONFIG_IFACE, Command(cmd, self)))

        if self.type == "bridge":
            old_ifnames = set(self.ifname)
            new_ifnames = set(new.ifname)

            # Remove interfaces that are not in the new bridge.
            for ifname in (old_ifnames - new_ifnames):
                commands.extend(self.removeFromBridge(ifname))

            # The bridge should not exist if this is the first physical interface
            # and bridge_empty is False.
            if len(self.ifname) == 0 and not self.bridge_empty:
                cmd = ["ip", "link", "delete", self.config_ifname]
                commands.append((-self.PRIO_CREATE_IFACE, Command(cmd, self)))

        return commands



"""
Node access rights

Operation             | owner | admin | trusted | guest
-------------------------------------------------------
install a chute       | yes   | yes   | limited | no
update a chute        | yes   | yes   | limited | no
remove a chute        | yes   | yes   | limited | no
change node settings  | yes   | yes   | no      | no
obtain SSH access     | yes   | yes   | no      | no

Trusted users have a limited ability to install, update, and remove chutes.
Trusted users can install any chute so long as it does not conflict with a
currently-installed chute. Trusted users can upgrade or remove chutes that
they have installed. These limitations are designed to allow multiple users
to access a shared node without allowing adverse interactions between users.
"""

def enforce_access_rights(update):
    """
    Check that the update should be allowed.
    """
    if update.old:
        if update.user != update.old.get_owner() and \
                update.user.role not in ["admin", "owner"]:
            raise Exception("Not permitted to modify existing chute {}".format(update.old.name))



import docker

from paradrop.base.exceptions import ChuteNotFound, ChuteNotRunning


class ChuteContainer(object):
    """
    Class for accessing information about a chute's container.
    """
    def __init__(self, name, docker_url="unix://var/run/docker.sock"):
        self.name = name
        self.docker_url = docker_url

    def getID(self):
        """
        Look up the container ID as used by Docker.
        """
        info = self.inspect()

        return info['Id']

    def getIP(self):
        """
        Look up the IP address assigned to the container.
        """
        info = self.inspect()

        if not info['State']['Running']:
            raise ChuteNotRunning("The chute is not running.")

        return info['NetworkSettings']['IPAddress']

    def getPID(self):
        """
        Look up the PID of the container, if running.
        """
        info = self.inspect()

        if not info['State']['Running']:
            raise ChuteNotRunning("The chute is not running.")

        return info['State']['Pid']

    def getPortConfiguration(self, port, protocol="tcp"):
        """
        Look up network port configuration.  This tells us if a port in the
        host is bound to a port inside the container.

        Returns a list, typically with zero or one elements.

        Example:

            [{
                "HostIp": "0.0.0.0",
                "HostPort": "32768"
            }]
        """
        info = self.inspect()
        key = "{}/{}".format(port, protocol)
        try:
            return info['NetworkSettings']['Ports'][key]
        except:
            return []

    def getStatus(self):
        """
        Return the status of the container (running, exited, paused).

        Returns "missing" if the chute does not exist.
        """
        try:
            info = self.inspect()
            return info['State']['Status']
        except ChuteNotFound:
            return "missing"

    def inspect(self):
        """
        Return the full container status from Docker.
        """
        client = docker.APIClient(base_url=self.docker_url, version='auto')
        try:
            info = client.inspect_container(self.name)
            return info
        except docker.errors.NotFound:
            raise ChuteNotFound("The chute could not be found.")

    def isRunning(self):
        """
        Check if container is running.

        Returns True/False; returns False if the container does not exist.
        """
        try:
            info = self.inspect()
            return info['State']['Running']
        except ChuteNotFound:
            return False



import os


def getAuthorizedKeys(user="paradrop"):
    homeDir = os.path.join("/home", user)
    if not os.path.isdir(homeDir):
        raise Exception("No home directory found for user {}".format(user))

    path = os.path.join(homeDir, ".ssh", "authorized_keys")
    if not os.path.isfile(path):
        return []

    keys = []
    with open(path, 'r') as source:
        for line in source:
            keys.append(line.strip())
    return keys


def writeAuthorizedKeys(keys, user="paradrop"):
    homeDir = os.path.join("/home", user)
    if not os.path.isdir(homeDir):
        raise Exception("No home directory found for user {}".format(user))

    sshDir = os.path.join(homeDir, ".ssh")
    if not os.path.exists(sshDir):
        os.mkdir(sshDir)

    path = os.path.join(sshDir, "authorized_keys")
    with open(path, "w") as output:
        for key in keys:
            output.write(key + "\n")


def addAuthorizedKey(key, user="paradrop"):
    keys = getAuthorizedKeys(user=user)

    for existing in keys:
        if key == existing:
            raise Exception("Key already exists")

    # One thing we avoid by rewriting the entire file is checking whether the
    # last line of the file had a newline or not.
    keys.append(key)
    writeAuthorizedKeys(keys, user=user)



from autobahn.twisted.websocket import WebSocketServerProtocol
from autobahn.twisted.websocket import WebSocketServerFactory

from paradrop.base.output import out

class AirsharkSpectrumProtocol(WebSocketServerProtocol):
    def __init__(self, factory):
        WebSocketServerProtocol.__init__(self)
        self.factory = factory

    def onOpen(self):
        out.info('ws /airshark/spectrum connected')
        self.factory.airshark_manager.add_spectrum_observer(self)

    def on_spectrum_data(self, data):
        self.sendMessage(data, True)

    def onClose(self, wasClean, code, reason):
        out.info('ws /airshark/spectrum disconnected: {}'.format(reason))
        self.factory.airshark_manager.remove_spectrum_observer(self)


class AirsharkSpectrumFactory(WebSocketServerFactory):
    def __init__(self, airshark_manager, *args, **kwargs):
        WebSocketServerFactory.__init__(self, *args, **kwargs)
        self.airshark_manager = airshark_manager

    def buildProtocol(self, addr):
        return AirsharkSpectrumProtocol(self)


class AirsharkAnalyzerProtocol(WebSocketServerProtocol):
    def __init__(self, factory):
        WebSocketServerProtocol.__init__(self)
        self.factory = factory

    def onOpen(self):
        out.info('ws /airshark/analyzer connected')
        self.factory.airshark_manager.add_analyzer_observer(self)

    def on_analyzer_message(self, message):
        # The message is in json format
        self.sendMessage(message, False)

    def onClose(self, wasClean, code, reason):
        out.info('ws /airshark/analyzer disconnected: {}'.format(reason))
        self.factory.airshark_manager.remove_analyzer_observer(self)


class AirsharkAnalyzerFactory(WebSocketServerFactory):
    def __init__(self, airshark_manager, *args, **kwargs):
        WebSocketServerFactory.__init__(self, *args, **kwargs)
        self.airshark_manager = airshark_manager

    def buildProtocol(self, addr):
        return AirsharkAnalyzerProtocol(self)



from paradrop.lib.utils import uci

from . import uciutils


def getOSWirelessConfig(update):
    """
    Read settings from networkInterfaces for wireless interfaces.
    Store wireless configuration settings in osWirelessConfig.
    """
    # old code under lib.internal.chs.chutelxc same function name

    interfaces = update.cache_get('networkInterfaces')
    if interfaces is None:
        return

    wifiIfaces = list()
    for iface in interfaces:
        # Only look at wifi interfaces.
        if not iface['type'].startswith("wifi"):
            continue

        config = {'type': 'wifi-iface'}
        options = {
            'device': iface['device'],
            'network': iface['externalIntf'],
            'mode': iface.get('mode', 'ap')
        }

        # Add wireless options.
        options.update(iface['wireless'])

        # Required for AP and client mode but not monitor mode.
        if 'ssid' in iface:
            options['ssid'] = iface['ssid']

        # Optional encryption settings
        if 'encryption' in iface:
            options['encryption'] = iface['encryption']
        if 'key' in iface:
            options['key'] = iface['key']

        wifiIfaces.append((config, options))

    update.cache_set('osWirelessConfig', wifiIfaces)


def setOSWirelessConfig(update):
    """
    Write settings from osWirelessConfig out to UCI files.
    """
    uciutils.setConfig(update,
            cacheKeys=['osWirelessConfig'],
            filepath=uci.getSystemPath("wireless"))


def revert_os_wireless_config(update):
    uciutils.setConfig(update,
            cacheKeys=['osWirelessConfig'],
            filepath=uci.getSystemPath("wireless"))



"""
Get information about network devices.

Endpoints for these functions can be found under /api/v1/network.
"""

import fnmatch
import json
import os

from klein import Klein

from paradrop.base import settings
from paradrop.lib.utils import parsing

from . import cors


def read_leases(path):
    """
    Read leases from a dnsmasq leases file.

    Returns a list of leases, each a dictionary containing the following fields.
    as_of: Time that lease information was last updated (seconds since Unix epoch).
    expires: DHCP expiration time (seconds since Unix epoch).
    mac_addr: MAC address of the device.
    ip_addr: IP address assigned to the device.
    hostname: Device hostname if reported.
    client_id: A client-specified identifier, which varies between devices.
    """
    # The format of the dnsmasq leases file is one entry per line with
    # space-separated fields.
    keys = ['expires', 'mac_addr', 'ip_addr', 'hostname', 'client_id']

    leases = []

    # Get mtime of the leases file, which gives a sense of the age.
    as_of = os.path.getmtime(path)

    with open(path, "r") as source:
        for line in source:
            parts = line.strip().split()
            entry = dict(zip(keys, parts))
            entry['as_of'] = as_of
            entry['expires'] = parsing.str_to_numeric(entry['expires'])

            # Note: I considered filtering out expired leases based on the
            # expiration time, but apparently dnsmasq leaves old expiration
            # times in this file even for active devices.
            leases.append(entry)

    return leases


def update_lease(leases, entry):
    """
    Update a dictionary of DHCP leases with a new entry.

    The dictionary should be indexed by MAC address. The new entry will be
    added to the dictionary unless it would replace an entry for the same MAC
    address from a more recent lease file.
    """
    mac_addr = entry['mac_addr']
    if mac_addr in leases:
        existing = leases[mac_addr]

        if existing['as_of'] >= entry['as_of']:
            return existing

    leases[mac_addr] = entry
    return entry


class NetworkApi(object):
    routes = Klein()

    def __init__(self):
        pass

    @routes.route("/devices", methods=["GET"])
    def get_devices(self, request):
        """
        List connected devices.
        """

        """
        Get detailed information about connected wireless stations.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/network/devices

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           [
             {
               "as_of": 1511806276,
               "client_id": "01:5c:59:48:7d:b9:e6",
               "expires": 1511816276,
               "ip_addr": "192.168.128.64",
               "mac_addr": "5c:59:48:7d:b9:e6",
               "hostname": "paradrops-iPod"
             }
           ]
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        leases = {}

        for root, dirs, files in os.walk(settings.RUNTIME_HOME_DIR):
            for file in fnmatch.filter(files, "dnsmasq-*.leases"):
                path = os.path.join(root, file)

                for entry in read_leases(path):
                    update_lease(leases, entry)

        return json.dumps(leases.values())



'''
Get system running status including CPU load, memory usage, network traffic.
'''
import time

import psutil
import six


class SystemStatus(object):
    INCLUDED_PARTITIONS = set(['/', '/writable'])

    def __init__(self):
        self.timestamp = time.time()
        self.cpu_load = []
        self.mem = dict(total = 0,
                        available = 0,
                        free = 0,
                        cached = 0,
                        buffers = 0)
        self.disk_partitions = {}
        self.network = {}

        partitions = psutil.disk_partitions()
        for p in partitions:
            if p.mountpoint in self.INCLUDED_PARTITIONS:
                usage = psutil.disk_usage(p.mountpoint)
                self.disk_partitions[p.mountpoint] = {
                    'total': usage.total,
                    'used': usage.used
                }


    def getStatus(self, max_age=0.8):
        """
        Get current system status.

        max_age: maximum tolerable age of cached status information.  Set to
        None to force a refresh regardless of cache age.

        Returns a dictionary with fields 'cpu_load', 'mem', 'disk', and
        'network'.
        """
        timestamp = time.time()
        if (max_age is None or timestamp > self.timestamp + max_age):
            self.timestamp = timestamp
            self.refreshCpuLoad()
            self.refreshMemoryInfo()
            self.refreshDiskInfo()
            self.refreshNetworkTraffic()

        result = {
            'cpu_load': self.cpu_load,
            'mem': self.mem,
            'disk': self.disk_partitions,
            'network': self.network
        }
        return result


    def refreshCpuLoad(self):
        self.cpu_load = list(map(int, psutil.cpu_percent(percpu=True)))


    def refreshMemoryInfo(self):
        mem = psutil.virtual_memory()
        self.mem['total'] = mem.total
        self.mem['available'] = mem.available
        self.mem['free'] = mem.free
        self.mem['cached'] = mem.cached
        self.mem['buffers'] = mem.buffers


    def refreshDiskInfo(self):
        for key, value in six.iteritems(self.disk_partitions):
            usage = psutil.disk_usage(key)
            self.disk_partitions[key]['total'] = usage.total
            self.disk_partitions[key]['used'] = usage.used


    def refreshNetworkTraffic(self):
        excluded_interfaces = set(["lo", 'br-lan', 'docker0', 'wlan0'])
        interfaces = {}

        stats = psutil.net_if_stats()
        for key, value in six.iteritems(stats):
            if key in excluded_interfaces:
                continue

            # ignore virtual ethernet interfaces
            if key.startswith('veth'):
                continue

            # Quick fix - ignore interfaces with dots because our server
            # refuses to accept keys with dots.
            if '.' in key:
                continue

            interfaces[key] = {
                'isup': value.isup,
                'speed': value.speed,
                'mtu': value.mtu
            }

        addresses = psutil.net_if_addrs()
        for key, value in six.iteritems(addresses):
            if key not in interfaces:
                continue

            for i in value:
                if i.family == 2:
                    interfaces[key]['ipv4'] = i.address
                    interfaces[key]['netmask'] = i.netmask
                elif i.family == 17:
                    interfaces[key]['mac'] = i.address

        traffic = psutil.net_io_counters(pernic=True)
        for key, value in six.iteritems(traffic):
            if key not in interfaces:
                continue

            interfaces[key]['bytes_sent'] = value.bytes_sent
            interfaces[key]['bytes_recv'] = value.bytes_recv
            interfaces[key]['packets_sent'] = value.packets_sent
            interfaces[key]['packets_recv'] = value.packets_recv
            interfaces[key]['errin'] = value.errin
            interfaces[key]['errout'] = value.errout
            interfaces[key]['dropin'] = value.dropin
            interfaces[key]['dropout'] = value.dropout

        self.network = interfaces

    @classmethod
    def getNetworkInfo(cls):
        interfaces = {}

        stats = psutil.net_if_stats()
        for key, value in six.iteritems(stats):
            interfaces[key] = value.__dict__
            interfaces[key]['addresses'] = []
            interfaces[key]['io_counters'] = None

        addresses = psutil.net_if_addrs()
        for key, value in six.iteritems(addresses):
            if key not in interfaces:
                continue

            for addr in value:
                interfaces[key]['addresses'].append(addr.__dict__)

        traffic = psutil.net_io_counters(pernic=True)
        for key, value in six.iteritems(traffic):
            if key not in interfaces:
                continue

            interfaces[key]['io_counters'] = value.__dict__

        return interfaces

    @classmethod
    def getProcessInfo(cls, pid):
        proc = psutil.Process(pid=pid)
        with proc.oneshot():
            proc_info = {
                'cpu_num': proc.cpu_num(),
                'cpu_times': proc.cpu_times().__dict__,
                'create_time': proc.create_time(),
                'memory_info': proc.memory_info().__dict__,
                'num_ctx_switches': proc.num_ctx_switches().__dict__,
                'num_threads': proc.num_threads()
            }
        return proc_info

    @classmethod
    def getSystemInfo(cls):
        system = {
            'boot_time': psutil.boot_time(),
            'cpu_count': psutil.cpu_count(),
            'cpu_stats': psutil.cpu_stats().__dict__,
            'cpu_times': [k.__dict__ for k in psutil.cpu_times(percpu=True)],
            'disk_io_counters': psutil.disk_io_counters().__dict__,
            'disk_usage': [],
            'net_io_counters': psutil.net_io_counters().__dict__,
            'swap_memory': psutil.swap_memory().__dict__,
            'virtual_memory': psutil.virtual_memory().__dict__
        }

        partitions = psutil.disk_partitions()
        for p in partitions:
            if p.mountpoint in cls.INCLUDED_PARTITIONS:
                usage = psutil.disk_usage(p.mountpoint)
                system['disk_usage'].append({
                    'mountpoint': p.mountpoint,
                    'total': usage.total,
                    'used': usage.used
                })

        return system



from __future__ import print_function
###################################################################
# Copyright 2013-2017 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

import sys

from paradrop.base import settings
from paradrop.lib.utils.pd_storage import PDStorage

from .chute import Chute


class ChuteStorage(PDStorage):

    """
        ChuteStorage class.

        This class holds onto the list of Chutes on this AP.

        It implements the PDStorage class which allows us to save the chuteList to disk transparently
    """
    # Class variable of chute list so all instances see the same thing
    chuteList = dict()

    def __init__(self, filename=None, save_timer=settings.FC_CHUTESTORAGE_SAVE_TIMER):
        if(not filename):
            filename = settings.FC_CHUTESTORAGE_FILE

        PDStorage.__init__(self, filename, save_timer)

        # Has it been loaded?
        if(len(ChuteStorage.chuteList) == 0):
            self.loadFromDisk()

    def setAttr(self, attr):
        """Save our attr however we want (as class variable for all to see)"""
        ChuteStorage.chuteList = attr

    def getAttr(self):
        """Get our attr (as class variable for all to see)"""
        return ChuteStorage.chuteList

    def getChuteList(self):
        """Return a list of the names of the chutes we know of."""
        return ChuteStorage.chuteList.values()

    def getChute(self, name):
        """Returns a reference to a chute we have in our cache, or None."""
        return ChuteStorage.chuteList.get(name, None)

    def deleteChute(self, ch):
        """Deletes a chute from the chute storage. Can be sent the chute object, or the chute name."""
        if (isinstance(ch, Chute)):
            del ChuteStorage.chuteList[ch.name]
        else:
            del ChuteStorage.chuteList[ch]
        self.saveToDisk()

    def saveChute(self, ch):
        """
            Saves the chute provided in our internal chuteList.
            Also since we just received a new chute to hold onto we should save our ChuteList to disk.
        """
        # check if there is a version of the chute already
        oldch = ChuteStorage.chuteList.get(ch.name, None)
        if(oldch != None):
            # we should merge these chutes so we don't lose any data
            oldch.__dict__.update(ch.__dict__)
            # TODO: do we need to deal with cache separate? Old code we did
        else:
            ChuteStorage.chuteList[ch.name] = ch

        self.saveToDisk()

    def clearChuteStorage(self):
        ChuteStorage.chuteList.clear()
        self.saveToDisk()

    #
    # Functions we override to implement PDStorage Properly
    #
    def attrSaveable(self):
        """Returns True if we should save the ChuteList, otherwise False."""
        return (type(ChuteStorage.chuteList) == dict)

    @classmethod
    def get_chute(cls, name):
        return cls.chuteList[name]


if(__name__ == '__main__'): # pragma: no cover
    def usage():
        print('Usage: $0 -ls : print chute storage details')
        exit(0)

    try:
        if(sys.argv[1] != '-ls'):
            usage()
    except Exception as e:
        print(e)
        usage()

    cs = ChuteStorage()

    chutes = cs.getChuteList()
    for ch in chutes:
        print(ch)



import json

from klein import Klein

from paradrop.base import pdutils
from . import cors


class ChangeApi(object):
    routes = Klein()

    def __init__(self, update_manager):
        self.update_manager = update_manager

    @routes.route('/', methods=['GET'])
    def get_changes(self, request):
        """
        Get list of active and queued changes.

        Note: we use the term "change" even though, internally, the objects are
        referred to as "updates". The word "update" has become so overloaded it
        causes much confusion. A "change" is an atomic and self-contained
        alteration to the running state of the system. A "change" could install
        a chute, remove a chute, change the host configuration, etc.
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        changes = []

        def dump_update(update, status):
            result = {
                'id': update.change_id,
                'updateClass': update.updateClass,
                'updateType': update.updateType,
                'user': update.user.__dict__,
                'name': getattr(update, 'name', None),
                'version': getattr(update, 'version', None),
                'status': status
            }
            return result

        for update in self.update_manager.active_changes.values():
            changes.append(dump_update(update, 'processing'))

        for update in self.update_manager.updateQueue:
            changes.append(dump_update(update, 'queued'))

        return json.dumps(changes)

    @routes.route('/', methods=['POST'])
    def create_change(self, request):
        """
        Schedule a new change.

        Note: we use the term "change" even though, internally, the objects are
        referred to as "updates". The word "update" has become so overloaded it
        causes much confusion. A "change" is an atomic and self-contained
        alteration to the running state of the system. A "change" could install
        a chute, remove a chute, change the host configuration, etc.
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        change = json.loads(request.content.read())
        change['tok'] = pdutils.timeint()

        # We will return the change ID to the caller for tracking and log
        # retrieval.
        change['change_id'] = self.update_manager.assign_change_id()

        self.update_manager.add_update(**change)

        result = {
            'change_id': change['change_id']
        }
        request.setHeader('Content-Type', 'application/json')
        return json.dumps(result)



from __future__ import print_function
from multiprocessing import Process, Value
from ctypes import c_bool
import os
import time

from .spectrum_reader import SpectrumReader

class Scanner(object):
    interface = None
    freqlist = None
    process = None
    debugfs_dir = None
    spectrum_reader = None

    def __init__(self, interface):
        self.interface = interface
        self.phy = ""
        self.debugfs_dir = self._find_debugfs_dir()
        if not self.debugfs_dir:
            raise Exception('Unable to access spectral_scan_ctl file for interface %s' % interface)

        self.ctl_file = '%s/spectral_scan_ctl' % self.debugfs_dir
        self.sample_count_file = '%s/spectral_count' % self.debugfs_dir
        self.short_repeat_file = '%s/spectral_short_repeat' % self.debugfs_dir
        self.cur_chan = 6
        self.sample_count = 8
        self.process = None
        self.set_freqs(2412, 2472, 5)
        self.spectrum_reader = SpectrumReader('%s/spectral_scan0' % self.debugfs_dir)

    def dev_to_phy(self, dev):
        f = open('/sys/class/net/%s/phy80211/name' % dev)
        phy = f.read().strip()
        f.close()
        return phy

    def _find_debugfs_dir(self):
        ''' search debugfs for spectral_scan_ctl for this interface '''
        for dirname, subd, files in os.walk('/sys/kernel/debug/ieee80211'):
            if 'spectral_scan_ctl' in files:
                phy = dirname.split(os.path.sep)[-2]
                if phy == self.dev_to_phy(self.interface):
                    self.phy = phy
                    return dirname
        return None

    def _scan(self, scanning):
        while scanning.value:
            cmd = 'iw dev %s scan trigger' % self.interface
            if self.freqlist:
                cmd = '%s freq %s' % (cmd, ' '.join(self.freqlist))
            os.system('%s >/dev/null 2>/dev/null' % cmd)
            time.sleep(0.1)

    def set_freqs(self, minf, maxf, spacing):
        self.freqlist = ['%s' % x for x in range(minf, maxf + spacing, spacing)]

    def cmd_chanscan(self):
        f = open(self.ctl_file, 'w')
        f.write("chanscan")
        f.close()

    def cmd_disable(self):
        f = open(self.ctl_file, 'w')
        f.write("disable")
        f.close()

    def cmd_set_samplecount(self, count):
        print("set sample count to %d" % count)
        f = open(self.sample_count_file, 'w')
        f.write("%s" % count)
        f.close()

    def cmd_set_short_repeat(self, short_repeat):
        f = open(self.short_repeat_file, 'w')
        f.write("%s" % short_repeat)
        f.close()

    def start(self):
        if self.process is None:
            self.scanning = Value(c_bool, True)
            self.process = Process(target=self._scan, args=(self.scanning,))
            self.process.start()

    def stop(self):
        self.cmd_disable()
        if self.process is not None:
            self.scanning.value = False
            self.process.join()
            self.process = None
            self.spectrum_reader.flush()

    def get_debugfs_dir(self):
        return self.debugfs_dir



"""
Utilities for reading from data structures.
"""

def getValue(struct, path, default=None):
    """
    Read a value from the data structure.

    Arguments:
    struct can comprise one or more levels of dicts and lists.
    path should be a string using dots to separate levels.
    default will be returned if the path cannot be traced.

    Example:
    getValue({'a': [1, 2, 3]}, "a.1") -> 2
    getValue({'a': [1, 2, 3]}, "a.3") -> None
    """
    parts = path.split(".")

    try:
        current = struct
        for part in parts:
            if isinstance(current, list):
                part = int(part)
            current = current[part]
        return current
    except:
        return default
